# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

# This module attempts to locate windows console histories.

import logging
import os
import struct
from typing import Tuple, Optional, Generator, Set, Dict, Any, Type, List

from volatility3.framework import interfaces, symbols, exceptions
from volatility3.framework import renderers
from volatility3.framework.interfaces import configuration
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import scanners
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows.extensions import pe, consoles
from volatility3.plugins.windows import pslist, info, verinfo
from volatility3.plugins.windows.registry import hivelist


vollog = logging.getLogger(__name__)


class Consoles(interfaces.plugins.PluginInterface):
    """Looks for Windows console buffers"""

    _required_framework_version = (2, 4, 0)

    # 2.0.0 - change the signature of `get_console_settings_from_registry`
    _version = (3, 0, 0)

    @classmethod
    def get_requirements(cls):
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="verinfo", component=verinfo.VerInfo, version=(1, 0, 0)
            ),
            requirements.VersionRequirement(
                name="info", component=info.Info, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="hivelist", component=hivelist.HiveList, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="bytes_scanner",
                component=scanners.BytesScanner,
                version=(1, 0, 0),
            ),
            requirements.BooleanRequirement(
                name="no_registry",
                description="Don't search the registry for possible values of CommandHistorySize and HistoryBufferMax",
                optional=True,
                default=False,
            ),
            requirements.ListRequirement(
                name="max_history",
                element_type=int,
                description="CommandHistorySize values to search for.",
                optional=True,
                default=[50],
            ),
            requirements.ListRequirement(
                name="max_buffers",
                element_type=int,
                description="HistoryBufferMax values to search for.",
                optional=True,
                default=[4],
            ),
        ]

    @classmethod
    def find_conhost_proc(
        cls, proc_list: Generator[interfaces.objects.ObjectInterface, None, None]
    ) -> Generator[Tuple[interfaces.objects.ObjectInterface, str], None, None]:
        """
        Walks the process list and returns the conhost instances.

        Args:
            proc_list: The process list generator

        Return:
            The process object and layer name for conhost
        """

        for proc in proc_list:
            if utility.array_to_string(proc.ImageFileName).lower() == "conhost.exe":
                proc_id = "Unknown"
                try:
                    proc_id = proc.UniqueProcessId
                    proc_layer_name = proc.add_process_layer()

                    yield proc, proc_layer_name

                except exceptions.InvalidAddressException as excp:
                    vollog.debug(
                        f"Process {proc_id}: invalid address {excp.invalid_address} in layer {excp.layer_name}"
                    )

    @classmethod
    def find_conhostexe(
        cls, conhost_proc: interfaces.objects.ObjectInterface
    ) -> Tuple[Optional[int], Optional[int]]:
        """
        Finds the base address of conhost.exe

        Args:
            conhost_proc: the process object for conhost.exe

        Returns:
            A tuple of:
            conhostexe_base: the base address of conhost.exe
            conhostexe_size: the size of the VAD for conhost.exe
        """
        for vad in conhost_proc.get_vad_root().traverse():
            filename = vad.get_file_name()
            if isinstance(filename, str) and filename.lower().endswith("conhost.exe"):
                base = vad.get_start()
                return base, vad.get_size()

        return None, None

    @classmethod
    def determine_conhost_version(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
        conhost_layer_name: str,
        conhost_base: int,
    ) -> Tuple[Optional[str], Dict[str, Type]]:
        """Tries to determine which symbol filename to use for the image's console information. This is similar to the
        netstat plugin.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            config_path: The config path where to find symbol files
            kernel_module_name: The name of the module for the kernel
            conhost_layer_name: The name of the conhot process memory layer
            conhost_base: the base address of conhost.exe

        Returns:
            The filename of the symbol table to use and the associated class types.
        """

        kernel = context.modules[kernel_module_name]

        is_64bit = symbols.symbol_table_is_64bit(
            context=context, symbol_table_name=kernel.symbol_table_name
        )

        if is_64bit:
            arch = "x64"
        else:
            arch = "x86"

        vers = info.Info.get_version_structure(context, kernel_module_name)

        kuser = info.Info.get_kuser_structure(context, kernel_module_name)

        try:
            vers_minor_version = int(vers.MinorVersion)
            nt_major_version = int(kuser.NtMajorVersion)
            nt_minor_version = int(kuser.NtMinorVersion)
        except ValueError:
            # vers struct exists, but is not an int anymore?
            raise NotImplementedError(
                "Kernel Debug Structure version format not supported!"
            )
        except Exception:
            # unsure what to raise here. Also, it might be useful to add some kind of fallback,
            # either to a user-provided version or to another method to determine conhost.exe's version
            raise exceptions.VolatilityException(
                "Kernel Debug Structure missing VERSION/KUSER structure, unable to determine Windows version!"
            )

        vollog.debug(
            f"Determined OS Version: {kuser.NtMajorVersion}.{kuser.NtMinorVersion} {vers.MajorVersion}.{vers.MinorVersion}"
        )

        if nt_major_version == 10 and arch == "x64":
            # win10 x64 has an additional class type we have to include.
            class_types = consoles.win10_x64_class_types
        else:
            # default to general class types
            class_types = consoles.class_types

        # these versions are listed explicitly because symbol files differ based on
        # version *and* architecture. this is currently the clearest way to show
        # the differences, even if it introduces a fair bit of redundancy.
        # furthermore, it is easy to append new versions.
        if arch == "x86":
            version_dict = {}
        else:
            version_dict = {
                (10, 0, 17763, 1): "consoles-win10-17763-x64",
                (10, 0, 17763, 3232): "consoles-win10-17763-3232-x64",
                (10, 0, 18362, 0): "consoles-win10-18362-x64",
                (10, 0, 19041, 0): "consoles-win10-19041-x64",
                (10, 0, 20348, 1): "consoles-win10-20348-x64",
                (10, 0, 20348, 1970): "consoles-win10-20348-1970-x64",
                (10, 0, 20348, 2461): "consoles-win10-20348-2461-x64",
                (10, 0, 20348, 2520): "consoles-win10-20348-2461-x64",
                (10, 0, 22000, 0): "consoles-win10-22000-x64",
                (10, 0, 22621, 1): "consoles-win10-22621-x64",
                (10, 0, 22621, 3527): "consoles-win10-22621-3527-x64",
                (10, 0, 25398, 0): "consoles-win10-22000-x64",
            }

        # we do not need to check for conhost's specific FileVersion in every case
        conhost_mod_version = 0  # keep it 0 as a default

        # we need to define additional version numbers (which are then found via conhost.exe's FileVersion header) in case there is
        # ambiguity _within_ an OS version. If such a version number (last number of the tuple) is defined for the current OS
        # we need to inspect conhost.exe's headers to see if we can grab the precise version
        if [
            (a, b, c, d)
            for a, b, c, d in version_dict
            if (a, b, c) == (nt_major_version, nt_minor_version, vers_minor_version)
            and d != 0
        ]:
            vollog.debug(
                "Requiring further version inspection due to OS version by checking conhost.exe's FileVersion header"
            )

            pe_table_name = intermed.IntermediateSymbolTable.create(
                context,
                configuration.path_join(config_path, "conhost"),
                "windows",
                "pe",
                class_types=pe.class_types,
            )

            try:
                (major, minor, product, build) = (
                    verinfo.VerInfo.get_version_information(
                        context, pe_table_name, conhost_layer_name, conhost_base
                    )
                )
                conhost_mod_version = build
                vollog.debug(
                    f"Found conhost.exe version {major}.{minor}.{product}.{build} in {conhost_layer_name} at base {conhost_base:#x}"
                )
            except (exceptions.InvalidAddressException, TypeError, AttributeError):
                # the following is IntelLayer specific and might need to be adapted to other architectures.
                physical_layer_name = context.layers[kernel.layer_name].config.get(
                    "memory_layer", None
                )
                if physical_layer_name:
                    ver = verinfo.VerInfo.find_version_info(
                        context, physical_layer_name, "CONHOST.EXE"
                    )

                    if ver:
                        conhost_mod_version = ver[3]
                        vollog.debug(
                            f"Determined conhost.exe's FileVersion: {conhost_mod_version}"
                        )
                    else:
                        vollog.debug("Could not determine conhost.exe's FileVersion.")
            else:
                vollog.debug(
                    "Unable to retrieve physical memory layer, skipping FileVersion check."
                )

        # when determining the symbol file we have to consider the following cases:
        # the determined version's symbol file is found by intermed.create -> proceed
        # the determined version's symbol file is not found by intermed -> intermed will throw an exc and abort
        # the determined version has no mapped symbol file -> if win10 use latest, otherwise throw exc
        # windows version cannot be determined -> throw exc

        filename = version_dict.get(
            (
                nt_major_version,
                nt_minor_version,
                vers_minor_version,
                conhost_mod_version,
            )
        )

        if not filename:
            # no match on filename means that we possibly have a version newer than those listed here.
            # try to grab the latest supported version of the current image NT version. If that symbol
            # version does not work, support has to be added manually.
            current_versions = [
                (nt_maj, nt_min, vers_min, conhost_ver)
                for nt_maj, nt_min, vers_min, conhost_ver in version_dict
                if nt_maj == nt_major_version
                and nt_min == nt_minor_version
                and vers_min <= vers_minor_version
                and conhost_ver <= conhost_mod_version
            ]
            current_versions.sort()

            if current_versions:
                latest_version = current_versions[-1]

                filename = version_dict.get(latest_version)

                vollog.debug(
                    f"Unable to find exact matching symbol file, going with latest: {filename}"
                )

            else:
                raise NotImplementedError(
                    f"This version of Windows is not supported: {nt_major_version}.{nt_minor_version} {vers.MajorVersion}.{vers_minor_version}!"
                )

        vollog.debug(f"Determined symbol filename: {filename}")

        return filename, class_types

    @classmethod
    def create_conhost_symbol_table(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
        conhost_layer_name: str,
        conhost_base: int,
    ) -> str:
        """Creates a symbol table for conhost structures.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            config_path: The config path where to find symbol files
            kernel_module_name: The name of the module of the kernel

        Returns:
            The name of the constructed symbol table
        """
        kernel = context.modules[kernel_module_name]

        table_mapping = {"nt_symbols": kernel.symbol_table_name}

        symbol_filename, class_types = cls.determine_conhost_version(
            context,
            config_path,
            kernel_module_name,
            conhost_layer_name,
            conhost_base,
        )

        if symbol_filename is None:
            raise ValueError(
                "Symbol filename could not be determined for conhost version"
            )

        vollog.debug(f"Using symbol file '{symbol_filename}' and types {class_types}")

        return intermed.IntermediateSymbolTable.create(
            context,
            configuration.path_join(config_path, "conhost"),
            os.path.join("windows", "consoles"),
            symbol_filename,
            class_types=class_types,
            table_mapping=table_mapping,
        )

    @classmethod
    def get_console_info(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
        procs: Generator[interfaces.objects.ObjectInterface, None, None],
        max_history: Set[int],
        max_buffers: Set[int],
    ) -> Generator[
        Tuple[
            interfaces.objects.ObjectInterface,
            Optional[interfaces.objects.ObjectInterface],
            List[Any],
        ],
        None,
        None,
    ]:
        """Gets the Console Information structure and its related properties for each conhost process

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            config_path: The config path where to find symbol files
            kernel_module_name: The name of the module for the kernel
            procs: list of process objects
            max_history: an initial set of CommandHistorySize values
            max_buffers: an initial list of HistoryBufferMax values

        Returns:
            The conhost process object, the console information structure, a dictionary of properties for
            that console information structure.
        """

        conhost_symbol_table = None

        for conhost_proc, proc_layer_name in cls.find_conhost_proc(procs):
            if not conhost_proc:
                vollog.info(
                    "Unable to find a valid conhost.exe process in the process list. Analysis cannot proceed."
                )
                continue
            vollog.debug(
                f"Found conhost process {conhost_proc} with pid {conhost_proc.UniqueProcessId}"
            )

            conhostexe_base, conhostexe_size = cls.find_conhostexe(conhost_proc)
            if not conhostexe_base:
                vollog.info(
                    "Unable to find the location of conhost.exe. Analysis cannot proceed."
                )
                continue
            if conhostexe_size is None:
                vollog.info(
                    "Unable to determine the size of conhost.exe.  Analysis cannot proceed."
                )
                continue
            vollog.debug(f"Found conhost.exe base at {conhostexe_base:#x}")

            proc_layer = context.layers[proc_layer_name]

            if conhost_symbol_table is None:
                conhost_symbol_table = cls.create_conhost_symbol_table(
                    context,
                    config_path,
                    kernel_module_name,
                    proc_layer_name,
                    conhostexe_base,
                )

            conhost_module = context.module(
                conhost_symbol_table, proc_layer_name, offset=conhostexe_base
            )

            found_console_info_for_proc = False
            console_info = None
            # scan for potential _CONSOLE_INFORMATION structures by using the CommandHistorySize
            for max_history_value in max_history:
                max_history_bytes = struct.pack("H", max_history_value)
                vollog.debug(
                    f"Scanning for CommandHistorySize value: {max_history_bytes}"
                )
                for address in proc_layer.scan(
                    context,
                    scanners.BytesScanner(max_history_bytes),
                    sections=[(conhostexe_base, conhostexe_size)],
                ):
                    console_info = None
                    console_properties = []

                    try:
                        console_info = conhost_module.object(
                            "_CONSOLE_INFORMATION",
                            offset=address
                            - conhost_module.get_type(
                                "_CONSOLE_INFORMATION"
                            ).relative_child_offset("CommandHistorySize"),
                            absolute=True,
                        )

                        if not any(
                            [
                                console_info.is_valid(max_buffer)
                                for max_buffer in max_buffers
                            ]
                        ):
                            continue

                        vollog.debug(
                            f"Getting Console Information properties for {console_info}"
                        )
                        console_properties.append(
                            {
                                "level": 0,
                                "name": "_CONSOLE_INFORMATION",
                                "address": console_info.vol.offset,
                                "data": "",
                            }
                        )
                        console_properties.append(
                            {
                                "level": 1,
                                "name": "_CONSOLE_INFORMATION.ScreenX",
                                "address": console_info.ScreenX.vol.offset,
                                "data": console_info.ScreenX,
                            }
                        )
                        console_properties.append(
                            {
                                "level": 1,
                                "name": "_CONSOLE_INFORMATION.ScreenY",
                                "address": console_info.ScreenY.vol.offset,
                                "data": console_info.ScreenY,
                            }
                        )
                        console_properties.append(
                            {
                                "level": 1,
                                "name": "_CONSOLE_INFORMATION.CommandHistorySize",
                                "address": console_info.CommandHistorySize.vol.offset,
                                "data": console_info.CommandHistorySize,
                            }
                        )
                        console_properties.append(
                            {
                                "level": 1,
                                "name": "_CONSOLE_INFORMATION.HistoryBufferCount",
                                "address": console_info.HistoryBufferCount.vol.offset,
                                "data": console_info.HistoryBufferCount,
                            }
                        )
                        console_properties.append(
                            {
                                "level": 1,
                                "name": "_CONSOLE_INFORMATION.HistoryBufferMax",
                                "address": console_info.HistoryBufferMax.vol.offset,
                                "data": console_info.HistoryBufferMax,
                            }
                        )
                        console_properties.append(
                            {
                                "level": 1,
                                "name": "_CONSOLE_INFORMATION.Title",
                                "address": console_info.Title.vol.offset,
                                "data": console_info.get_title(),
                            }
                        )
                        console_properties.append(
                            {
                                "level": 1,
                                "name": "_CONSOLE_INFORMATION.OriginalTitle",
                                "address": console_info.OriginalTitle.vol.offset,
                                "data": console_info.get_original_title(),
                            }
                        )

                        vollog.debug(
                            f"Getting ConsoleProcessList entries for {console_info.ConsoleProcessList}"
                        )
                        console_properties.append(
                            {
                                "level": 1,
                                "name": "_CONSOLE_INFORMATION.ProcessCount",
                                "address": console_info.ProcessCount.vol.offset,
                                "data": console_info.ProcessCount,
                            }
                        )
                        console_properties.append(
                            {
                                "level": 1,
                                "name": "_CONSOLE_INFORMATION.ConsoleProcessList",
                                "address": console_info.ConsoleProcessList.vol.offset,
                                "data": "",
                            }
                        )
                        for index, attached_proc in enumerate(
                            console_info.get_processes()
                        ):
                            console_properties.append(
                                {
                                    "level": 2,
                                    "name": f"_CONSOLE_INFORMATION.ConsoleProcessList.ConsoleProcess_{index}",
                                    "address": attached_proc.ConsoleProcess.dereference().vol.offset,
                                    "data": "",
                                }
                            )
                            console_properties.append(
                                {
                                    "level": 2,
                                    "name": f"_CONSOLE_INFORMATION.ConsoleProcessList.ConsoleProcess_{index}_ProcessId",
                                    "address": attached_proc.ConsoleProcess.ProcessId.vol.offset,
                                    "data": attached_proc.ConsoleProcess.ProcessId,
                                }
                            )
                            console_properties.append(
                                {
                                    "level": 2,
                                    "name": f"_CONSOLE_INFORMATION.ConsoleProcessList.ConsoleProcess_{index}_ProcessHandle",
                                    "address": attached_proc.ConsoleProcess.ProcessHandle.vol.offset,
                                    "data": hex(
                                        attached_proc.ConsoleProcess.ProcessHandle
                                    ),
                                }
                            )

                        vollog.debug(
                            f"Getting ExeAliasList entries for {console_info.ExeAliasList}"
                        )
                        console_properties.append(
                            {
                                "level": 1,
                                "name": "_CONSOLE_INFORMATION.ExeAliasList",
                                "address": console_info.ExeAliasList.vol.offset,
                                "data": "",
                            }
                        )
                        if console_info.ExeAliasList:
                            for index, exe_alias_list in enumerate(
                                console_info.get_exe_aliases()
                            ):
                                try:
                                    console_properties.append(
                                        {
                                            "level": 2,
                                            "name": f"_CONSOLE_INFORMATION.ExeAliasList.AliasList_{index}",
                                            "address": exe_alias_list.vol.offset,
                                            "data": "",
                                        }
                                    )
                                    console_properties.append(
                                        {
                                            "level": 2,
                                            "name": f"_CONSOLE_INFORMATION.ExeAliasList.AliasList_{index}.ExeName",
                                            "address": exe_alias_list.ExeName.vol.offset,
                                            "data": exe_alias_list.get_exename(),
                                        }
                                    )
                                    for alias_index, alias in enumerate(
                                        exe_alias_list.get_aliases()
                                    ):
                                        console_properties.append(
                                            {
                                                "level": 3,
                                                "name": f"_CONSOLE_INFORMATION.ExeAliasList.AliasList_{index}.Alias_{alias_index}.Source",
                                                "address": alias.Source.vol.offset,
                                                "data": alias.get_source(),
                                            }
                                        )
                                        console_properties.append(
                                            {
                                                "level": 3,
                                                "name": f"_CONSOLE_INFORMATION.ExeAliasList.AliasList_{index}.Alias_{alias_index}.Target",
                                                "address": alias.Target.vol.offset,
                                                "data": alias.get_target(),
                                            }
                                        )
                                except Exception as e:
                                    vollog.debug(
                                        f"reading {exe_alias_list} encountered exception {e}"
                                    )

                        vollog.debug(
                            f"Getting HistoryList entries for {console_info.HistoryList}"
                        )
                        console_properties.append(
                            {
                                "level": 1,
                                "name": "_CONSOLE_INFORMATION.HistoryList",
                                "address": console_info.HistoryList.vol.offset,
                                "data": "",
                            }
                        )
                        for index, command_history in enumerate(
                            console_info.get_histories()
                        ):
                            try:
                                console_properties.append(
                                    {
                                        "level": 2,
                                        "name": f"_CONSOLE_INFORMATION.HistoryList.CommandHistory_{index}",
                                        "address": command_history.vol.offset,
                                        "data": "",
                                    }
                                )
                                console_properties.append(
                                    {
                                        "level": 2,
                                        "name": f"_CONSOLE_INFORMATION.HistoryList.CommandHistory_{index}_Application",
                                        "address": command_history.Application.vol.offset,
                                        "data": command_history.get_application(),
                                    }
                                )
                                console_properties.append(
                                    {
                                        "level": 2,
                                        "name": f"_CONSOLE_INFORMATION.HistoryList.CommandHistory_{index}_ProcessHandle",
                                        "address": command_history.ConsoleProcessHandle.ProcessHandle.vol.offset,
                                        "data": hex(
                                            command_history.ConsoleProcessHandle.ProcessHandle
                                        ),
                                    }
                                )
                                console_properties.append(
                                    {
                                        "level": 2,
                                        "name": f"_CONSOLE_INFORMATION.HistoryList.CommandHistory_{index}_CommandCount",
                                        "address": None,
                                        "data": command_history.CommandCount,
                                    }
                                )
                                console_properties.append(
                                    {
                                        "level": 2,
                                        "name": f"_CONSOLE_INFORMATION.HistoryList.CommandHistory_{index}_LastDisplayed",
                                        "address": command_history.LastDisplayed.vol.offset,
                                        "data": command_history.LastDisplayed,
                                    }
                                )
                                for (
                                    cmd_index,
                                    bucket_cmd,
                                ) in command_history.get_commands():
                                    try:
                                        console_properties.append(
                                            {
                                                "level": 3,
                                                "name": f"_CONSOLE_INFORMATION.HistoryList.CommandHistory_{index}_Command_{cmd_index}",
                                                "address": bucket_cmd.vol.offset,
                                                "data": bucket_cmd.get_command_string(),
                                            }
                                        )
                                    except Exception as e:
                                        vollog.debug(
                                            f"reading {bucket_cmd} encountered exception {e}"
                                        )
                            except Exception as e:
                                vollog.debug(
                                    f"reading {command_history} encountered exception {e}"
                                )

                        try:
                            vollog.debug(
                                f"Getting ScreenBuffer entries for {console_info}"
                            )
                            console_properties.append(
                                {
                                    "level": 1,
                                    "name": "_CONSOLE_INFORMATION.CurrentScreenBuffer",
                                    "address": console_info.CurrentScreenBuffer.vol.offset,
                                    "data": "",
                                }
                            )
                            for screen_index, screen_info in enumerate(
                                console_info.get_screens()
                            ):
                                try:
                                    console_properties.append(
                                        {
                                            "level": 2,
                                            "name": f"_CONSOLE_INFORMATION.ScreenBuffer_{screen_index}",
                                            "address": screen_info,
                                            "data": "",
                                        }
                                    )
                                    console_properties.append(
                                        {
                                            "level": 2,
                                            "name": f"_CONSOLE_INFORMATION.ScreenBuffer_{screen_index}.ScreenX",
                                            "address": None,
                                            "data": screen_info.ScreenX,
                                        }
                                    )
                                    console_properties.append(
                                        {
                                            "level": 2,
                                            "name": f"_CONSOLE_INFORMATION.ScreenBuffer_{screen_index}.ScreenY",
                                            "address": None,
                                            "data": screen_info.ScreenY,
                                        }
                                    )
                                    console_properties.append(
                                        {
                                            "level": 2,
                                            "name": f"_CONSOLE_INFORMATION.ScreenBuffer_{screen_index}.Dump",
                                            "address": None,
                                            "data": "\n".join(screen_info.get_buffer()),
                                        }
                                    )
                                except Exception as e:
                                    vollog.debug(
                                        f"reading {screen_info} encountered exception {e}"
                                    )
                        except Exception as e:
                            vollog.debug(
                                f"reading _CONSOLE_INFORMATION.CurrentScreenBuffer encountered exception {e}"
                            )

                    except exceptions.PagedInvalidAddressException as exp:
                        vollog.debug(
                            f"Required memory at {exp.invalid_address:#x} is not valid"
                        )
                        continue

                    if console_info and console_properties:
                        found_console_info_for_proc = True
                        yield conhost_proc, console_info, console_properties

            if not found_console_info_for_proc:
                yield conhost_proc, console_info or None, []

    @classmethod
    def get_console_settings_from_registry(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
        max_history: Set[int],
        max_buffers: Set[int],
    ) -> Tuple[Set[int], Set[int]]:
        """
        Walks the Registry user hives and extracts any CommandHistorySize and HistoryBufferMax values
        for scanning

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            config_path: The config path where to find symbol files
            kernel_module_name: The name of the module for the kernel
            max_history: an initial set of CommandHistorySize values
            max_buffers: an initial list of HistoryBufferMax values

        Returns:
            The updated max_history and max_buffers sets.
        """
        vollog.debug(
            f"Possible CommandHistorySize values before checking Registry: {max_history}"
        )
        vollog.debug(
            f"Possible HistoryBufferMax values before checking Registry: {max_buffers}"
        )

        for hive in hivelist.HiveList.list_hives(
            context=context,
            base_config_path=config_path,
            kernel_module_name=kernel_module_name,
            hive_offsets=None,
        ):
            try:
                for value in hive.get_key("Console").get_values():
                    val_name = str(value.get_name())
                    if val_name == "HistoryBufferSize":
                        max_history.add(value.decode_data())
                    elif val_name == "NumberOfHistoryBuffers":
                        max_buffers.add(value.decode_data())
            except Exception:
                continue

        return max_history, max_buffers

    def _generator(
        self, procs: Generator[interfaces.objects.ObjectInterface, None, None]
    ):
        """
        Generates the console information to use in rendering

        Args:
            procs: the process list filtered to conhost.exe instances
        """

        max_history = set(self.config.get("max_history", [50]))
        max_buffers = set(self.config.get("max_buffers", [4]))
        no_registry = self.config.get("no_registry")

        if no_registry is False:
            max_history, max_buffers = self.get_console_settings_from_registry(
                self.context,
                self.config_path,
                self.config["kernel"],
                max_history,
                max_buffers,
            )

        vollog.debug(f"Possible CommandHistorySize values: {max_history}")
        vollog.debug(f"Possible HistoryBufferMax values: {max_buffers}")

        proc = None
        for proc, console_info, console_properties in self.get_console_info(
            self.context,
            self.config_path,
            self.config["kernel"],
            procs,
            max_history,
            max_buffers,
        ):
            process_name = utility.array_to_string(proc.ImageFileName)
            process_pid = proc.UniqueProcessId

            if console_info and console_properties:
                for console_property in console_properties:
                    yield (
                        console_property["level"],
                        (
                            process_pid,
                            process_name,
                            format_hints.Hex(console_info.vol.offset),
                            console_property["name"],
                            (
                                renderers.NotApplicableValue()
                                if console_property["address"] is None
                                else format_hints.Hex(console_property["address"])
                            ),
                            (
                                str(console_property["data"])
                                if console_property["data"]
                                else renderers.NotAvailableValue()
                            ),
                        ),
                    )
            else:
                yield (
                    0,
                    (
                        process_pid,
                        process_name,
                        (
                            format_hints.Hex(console_info.vol.offset)
                            if console_info
                            else renderers.NotApplicableValue()
                        ),
                        "_CONSOLE_INFORMATION",
                        renderers.NotApplicableValue(),
                        "Console Information Not Found",
                    ),
                )

        if proc is None:
            vollog.warn("No conhost.exe processes found.")

    def _conhost_proc_filter(self, proc: interfaces.objects.ObjectInterface) -> bool:
        """
        Used to filter to only conhost.exe processes
        """
        process_name = utility.array_to_string(proc.ImageFileName)

        return process_name.lower() != "conhost.exe"

    def run(self):
        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("ConsoleInfo", format_hints.Hex),
                ("Property", str),
                ("Address", format_hints.Hex),
                ("Data", str),
            ],
            self._generator(
                pslist.PsList.list_processes(
                    context=self.context,
                    kernel_module_name=self.config["kernel"],
                    filter_func=self._conhost_proc_filter,
                )
            ),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import re
import logging
from typing import Optional, Union

from volatility3.framework import interfaces, objects, constants, exceptions

vollog = logging.getLogger(__name__)


def rol(value: int, count: int, max_bits: int = 64) -> int:
    """A rotate-left instruction in Python"""
    max_bits_mask = (1 << max_bits) - 1
    return (value << count % max_bits) & max_bits_mask | (
        (value & max_bits_mask) >> (max_bits - (count % max_bits))
    )


def bswap_32(value: int) -> int:
    value = ((value << 8) & 0xFF00FF00) | ((value >> 8) & 0x00FF00FF)

    return ((value << 16) | (value >> 16)) & 0xFFFFFFFF


def bswap_64(value: int) -> int:
    low = bswap_32(value >> 32)
    high = bswap_32(value & 0xFFFFFFFF)

    return ((high << 32) | low) & 0xFFFFFFFFFFFFFFFF


def array_to_string(
    array: "objects.Array",
    count: Optional[int] = None,
    errors: str = "replace",
    block_size=32,
    encoding="utf-8",
) -> str:
    """Takes a Volatility 'Array' of characters and returns a Python string.

    Args:
        array: The Volatility `Array` object containing character elements.
        count: Optional maximum number of characters to convert. If None, the function
               processes the entire array.
        errors: Specifies error handling behavior for decoding, defaulting to "replace".
        block_size: Reading block size. Defaults to 32

    Returns:
        A decoded string representation of the character array.
    """
    # TODO: Consider checking the Array's target is a native char
    if not isinstance(array, objects.Array):
        raise TypeError("Array_to_string takes an Array of char")

    if count is None:
        count = array.vol.count

    return address_to_string(
        context=array._context,
        layer_name=array.vol.layer_name,
        address=array.vol.offset,
        count=count,
        errors=errors,
        block_size=block_size,
        encoding=encoding,
    )


def pointer_to_string(
    pointer: "objects.Pointer",
    count: int,
    errors: str = "replace",
    block_size=32,
    encoding="utf-8",
) -> str:
    """Takes a Volatility 'Pointer' to characters and returns a Python string.

    Args:
        pointer: A `Pointer` object containing character elements.
        count: Optional maximum number of characters to convert. If None, the function
               processes the entire array.
        errors: Specifies error handling behavior for decoding, defaulting to "replace".
        block_size: Reading block size. Defaults to 32

    Returns:
        A decoded string representation of the data referenced by the pointer.
    """
    if not isinstance(pointer, objects.Pointer):
        raise TypeError("pointer_to_string takes a Pointer")

    if count < 1:
        raise ValueError("pointer_to_string requires a positive count")

    return address_to_string(
        context=pointer._context,
        layer_name=pointer.vol.layer_name,
        address=pointer,
        count=count,
        errors=errors,
        block_size=block_size,
        encoding=encoding,
    )


def gather_contiguous_bytes_from_address(
    context, data_layer, starting_address: int, count: int
) -> bytes:
    """
    This method reconstructs a string from memory while also carefully examining each page

    It goes page-by-page reading the bytes. This is done by calculating page boundaries
    and then only reading one page at a time.

    If a page is missing, the code initially catches the exception.
    If data is non-empty (meaning at least one read succeeded), then we return what was read
    If the first page fails, then we re-raise the exception
    """

    data = b""

    if isinstance(data_layer, interfaces.layers.TranslationLayerInterface):
        last_address = starting_address

        for address, length, _, _, _ in data_layer.mapping(
            offset=starting_address, length=count, ignore_errors=True
        ):
            # we hit a swapped out page
            if last_address != address:
                break

            data += data_layer.read(address, length)

            last_address = address + length

    elif starting_address + count < data_layer.maximum_address:
        data = data_layer.read(starting_address, count)

    # if we were able to read from the first page, we want to try and construct the string
    # if the first page fails -> throw exception
    if data:
        return data
    else:
        raise exceptions.InvalidAddressException(
            layer_name=data_layer, invalid_address=starting_address
        )


def bytes_to_decoded_string(
    data: bytes, encoding: str, errors: str, return_truncated: bool = True
) -> str:
    """
    Args:
        data: The `bytes` buffer containing the string of a string at offset 0
        encoding: An encoding value for the encoding parameter of `bytes.decode`
        errors: An errors value for the errors parameter of `bytes.decode`
        return_truncated: Dictates whether truncated strings should be returned or
        if a ValueError should be thrown if a truncated (broken) string was decoded
    Returns:
        bytes: The decoded string starting at offset of data

    This function takes a bytes buffer that contains at a string of unknown
    length starting at the first byte, and returns the properly decoded string

    It starts by using Python's `bytes.decode` to attempt to decode the entire string
    It then finds the termination character (\ufffd or \x00) and splices the string
    Finally, it returns this spliced string after its been decoded with the
        caller-specified encoding
    """
    # this is the standard byte used to replace bad unicode characters
    unicode_replacement_char = "\ufffd"

    # used to find the terminating byte
    termination_re = re.compile(f"{unicode_replacement_char}|\x00")

    # run over the entire string, letting Python replace invalid characters
    full_decoded_string = data.decode(encoding=encoding, errors="replace")

    # stop at the first terminating character or get the whole string if not found
    try:
        idx = termination_re.search(full_decoded_string).start()
    except AttributeError:
        if return_truncated:
            idx = len(full_decoded_string)
        else:
            raise ValueError(
                "return_truncated set to False and truncated string decoded."
            )

    # cut at terminating byte, if found
    data = bytes(full_decoded_string[:idx], encoding=encoding)

    # return with caller-specified encoding and errors
    return data.decode(encoding=encoding, errors=errors)


def address_to_string(
    context: interfaces.context.ContextInterface,
    layer_name: str,
    address: int,
    count: int,
    errors: str = "replace",
    block_size=32,
    encoding="utf-8",
) -> str:
    """Reads a null-terminated string from a given specified memory address, processing
       it in blocks for efficiency.

    Args:
        context: The context used to retrieve memory layers and symbol tables
        layer_name: The name of the memory layer to read from
        address: The address where the string is located in memory
        count: The number of bytes to read
        errors: The error handling scheme to use for encoding errors. Defaults to "replace"
        block_size: Reading block size. Defaults to 32

    Returns:
        The decoded string extracted from memory.
    """
    if not isinstance(address, int):
        raise TypeError("Address must be a valid integer")

    if count < 1:
        raise ValueError("Count must be greater than 0")

    layer = context.layers[layer_name]

    data = gather_contiguous_bytes_from_address(context, layer, address, count)

    return bytes_to_decoded_string(data=data, errors=errors, encoding=encoding)


def array_of_pointers(
    array: interfaces.objects.ObjectInterface,
    count: int,
    subtype: Union[str, interfaces.objects.Template],
    context: interfaces.context.ContextInterface,
) -> interfaces.objects.ObjectInterface:
    """Takes an object, and recasts it as an array of pointers to subtype."""
    symbol_table = array.vol.type_name.split(constants.BANG)[0]
    if isinstance(subtype, str) and context is not None:
        subtype = context.symbol_space.get_type(subtype)
    if not isinstance(subtype, interfaces.objects.Template) or subtype is None:
        raise TypeError(
            "Subtype must be a valid template (or string name of an object template)"
        )
    # We have to clone the pointer class, or we'll be defining the pointer subtype for all future pointers
    subtype_pointer = context.symbol_space.get_type(
        symbol_table + constants.BANG + "pointer"
    ).clone()
    subtype_pointer.update_vol(subtype=subtype)
    return array.cast("array", count=count, subtype=subtype_pointer)


def dynamically_sized_array_of_pointers(
    context: interfaces.context.ContextInterface,
    array: interfaces.objects.ObjectInterface,
    subtype: Union[str, interfaces.objects.Template],
    iterator_guard_value: int,
    stop_value: int = 0,
    stop_on_invalid_pointers: bool = True,
) -> interfaces.objects.ObjectInterface:
    """Iterates over a dynamically sized array of pointers (e.g. NULL-terminated).
    Array iteration should always be performed with an arbitrary guard value as maximum size,
    to prevent running forever in case something unexpected happens.

        Args:
            context: The context on which to operate.
            array: The object to cast to an array.
            iterator_guard_value: Stop iterating when the iterator index is greater than this value. This is an extra-safety against smearing.
            subtype: The subtype of the array's pointers.
            stop_value: Stop value used to determine when to terminate iteration once it is encountered. Defaults to 0 (NULL-terminated arrays).
            stop_on_invalid_pointers: Determines whether to stop iterating or not when an invalid pointer is encountered. This can be useful for arrays
    that are known to have smeared entries before the end.

        Returns:
            An array of pointer objects
    """
    new_count = 0
    sym_table_name = array.get_symbol_table_name()
    sym_table = context.symbol_space[sym_table_name]
    ptr_size = sym_table.get_type("pointer").size
    layer_name = array.vol.layer_name

    offset = array.vol.offset
    entry = None
    while entry != stop_value and new_count < iterator_guard_value:
        try:
            entry = context.object(
                sym_table_name + constants.BANG + "pointer",
                offset=offset,
                layer_name=layer_name,
            )
        except exceptions.InvalidAddressException:
            break

        if not entry.is_readable() and stop_on_invalid_pointers:
            break

        offset += ptr_size
        new_count += 1
    else:
        vollog.log(
            constants.LOGLEVEL_V,
            f"""Iterator guard value {iterator_guard_value} reached while iterating over array at offset {array.vol.offset:#x}.\
 This means that there is a bug (e.g. smearing) with this array, or that it may contain valid entries past the iterator guard value.""",
        )

    # Leverage the "Array" object instead of returning a Python list
    return array_of_pointers(
        array=array, count=new_count, subtype=subtype, context=context
    )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import bz2
import contextlib
import gzip
import hashlib
import logging
import lzma
import os
import ssl
import sys
import urllib.parse
import urllib.request
import zipfile
from typing import Any, IO, List, Optional
from urllib import error

from volatility3 import framework
from volatility3.framework import constants, exceptions

try:
    import magic

    HAS_MAGIC = True
except ImportError:
    HAS_MAGIC = False

try:
    # Import so that the handler is found by the framework.class_subclasses callc
    from smb import SMBHandler as SMBHandler  # lgtm [py/unused-import]
except ImportError:
    # If we fail to import this, it means that SMB handling won't be available
    pass

vollog = logging.getLogger(__name__)


# TODO: Type-annotating the ResourceAccessor.open method is difficult because HTTPResponse is not actually an IO[Any] type
#   fix this


def cascadeCloseFile(new_fp: IO[bytes], original_fp: IO[bytes]) -> IO[bytes]:
    """Really horrible solution for ensuring files aren't left open

    Args:
        new_fp: The file pointer constructed based on the original file pointer
        original_fp: The original file pointer that should be closed when the new file pointer is closed, but isn't
    """

    def close():
        original_fp.close()
        return new_fp.__class__.close(new_fp)

    new_fp.close = close
    return new_fp


class ResourceAccessor:
    """Object for opening URLs as files (downloading locally first if
    necessary)"""

    list_handlers = True

    def __init__(
        self,
        progress_callback: Optional[constants.ProgressCallback] = None,
        context: Optional[ssl.SSLContext] = None,
        enable_cache: bool = True,
    ) -> None:
        """Creates a resource accessor.

        Note: context is an SSL context, not a volatility context
        """
        self._progress_callback = progress_callback
        self._context = context
        self._handlers = list(framework.class_subclasses(urllib.request.BaseHandler))
        self._enable_cache = enable_cache
        if self.list_handlers:
            vollog.log(
                constants.LOGLEVEL_VVV,
                f"Available URL handlers: {', '.join([x.__name__ for x in self._handlers])}",
            )
            self.__class__.list_handlers = False

    def uses_cache(self, url: str) -> bool:
        """Determines whether a URLs contents should be cached"""
        parsed_url = urllib.parse.urlparse(url)

        return (
            self._enable_cache and parsed_url.scheme not in self._non_cached_schemes()
        )

    @staticmethod
    def _non_cached_schemes() -> List[str]:
        """Returns the list of schemes not to be cached"""
        result = ["file"]
        for clazz in framework.class_subclasses(VolatilityHandler):
            result += clazz.non_cached_schemes()
        return result

    # Current urllib.request.urlopen returns Any, so we do the same
    def open(self, url: str, mode: str = "rb") -> Any:
        """Returns a file-like object for a particular URL opened in mode.

        If the file is remote, it will be downloaded and locally cached
        """
        urllib.request.install_opener(urllib.request.build_opener(*self._handlers))

        # Python bug 46654
        if sys.platform == "win32":
            # We only need to worry about UNC paths on windows, on linux they'd be smb:// and need pysmb or similar
            parsed_url = urllib.parse.urlparse(url, scheme="file")
            # Only worry about file scheme URLs, make sure that there's either a host or
            # the unparsing left an extra slash at the start (which will get lost with urlunparse)
            if parsed_url.scheme == "file" and (
                parsed_url.netloc or parsed_url.path.startswith("//")
            ):
                # Change the netloc to '/' and then prepend the netloc to the path
                # Urlunparse will remove extra initial slashes from path, hence setting netloc
                new_url = urllib.parse.urlunparse(
                    (
                        parsed_url.scheme,
                        "/",
                        "/" + parsed_url.netloc + parsed_url.path,
                        parsed_url.params,
                        parsed_url.query,
                        parsed_url.fragment,
                    )
                )
                vollog.log(
                    constants.LOGLEVEL_VVVV,
                    f"UNC path detected, converted path {url} to {new_url}",
                )
                url = new_url

        try:
            fp = urllib.request.urlopen(url, context=self._context)
        except error.URLError as excp:
            if excp.args:
                if isinstance(excp.args[0], ssl.SSLCertVerificationError):
                    vollog.warning(
                        "SSL certificate verification failed: attempting UNVERIFIED retrieval"
                    )
                    non_verifying_ctx = ssl.SSLContext()
                    non_verifying_ctx.check_hostname = False
                    non_verifying_ctx.verify_mode = ssl.CERT_NONE
                    fp = urllib.request.urlopen(url, context=non_verifying_ctx)
                else:
                    raise excp
            else:
                raise excp
        except ValueError as excp:
            # Reraise errors such as proxy auth errors as offline exception errors
            # Example Proxy auth error - ValueError: AbstractDigestAuthHandler does not support the following scheme: 'Negotiate'
            vollog.info(f"Cannot access {url} due to {excp} - Setting OFFLINE")
            constants.OFFLINE = True
            raise exceptions.OfflineException(url)
        except exceptions.OfflineException:
            vollog.info(f"Not accessing {url} in offline mode")
            raise

        with contextlib.closing(fp) as fp:
            # Cache the file locally

            if not self.uses_cache(url):
                # ZipExtFiles (files in zips) cannot seek, so must be cached in order to use and/or decompress
                curfile = urllib.request.urlopen(url, context=self._context)
            else:
                # TODO: find a way to check if we already have this file (look at http headers?)
                block_size = 1028 * 8
                temp_filename = os.path.join(
                    constants.CACHE_PATH,
                    "data_"
                    + hashlib.sha512(bytes(url, "raw_unicode_escape")).hexdigest()
                    + ".cache",
                )

                try:
                    content_length = int(fp.info().get("Content-Length", -1))
                except (AttributeError, ValueError):
                    # If our fp doesn't have an info member, carry on gracefully
                    content_length = -1

                if not os.path.exists(temp_filename):
                    vollog.debug(f"Caching file at: {temp_filename}")
                    cache_file_size = -1

                    try:
                        with open(temp_filename, "wb") as cache_file:
                            count = 0
                            block = fp.read(block_size)
                            while block:
                                count += len(block)
                                if self._progress_callback:
                                    self._progress_callback(
                                        count * 100 / max(count, int(content_length)),
                                        f"Reading file {url}",
                                    )
                                cache_file.write(block)
                                block = fp.read(block_size)
                            cache_file.seek(0, os.SEEK_END)
                            cache_file_size = cache_file.tell()
                    finally:
                        if cache_file_size < content_length:
                            os.remove(temp_filename)
                            raise ValueError("Cached file did not download completely")
                else:
                    vollog.debug(
                        f"Trying to use already cached file at: {temp_filename}"
                    )

                # Re-open the cache with a different mode
                # Since we don't want people thinking they're able to save to the cache file,
                # open it in read mode only and allow breakages to happen if they wanted to write
                curfile = open(temp_filename, mode="rb")

        # Validate the hash or delete the temp_filename and report an error

        # Determine whether the file is a particular type of file, and if so, open it as such
        IMPORTED_MAGIC = False
        if HAS_MAGIC:
            stop = False
            while not stop:
                detected = None
                with contextlib.suppress(AttributeError, IOError):
                    # Detect the content
                    detected = magic.detect_from_fobj(curfile)
                    IMPORTED_MAGIC = True
                    # This is because python-magic and file provide a magic module
                    # Only file's python has magic.detect_from_fobj

                if detected:
                    if detected.mime_type == "application/x-xz":
                        curfile = cascadeCloseFile(
                            lzma.LZMAFile(curfile, mode), curfile
                        )
                    elif detected.mime_type == "application/x-bzip2":
                        curfile = cascadeCloseFile(bz2.BZ2File(curfile, mode), curfile)
                    elif detected.mime_type == "application/x-gzip":
                        curfile = cascadeCloseFile(
                            gzip.GzipFile(fileobj=curfile, mode=mode), curfile
                        )
                    if detected.mime_type in [
                        "application/x-xz",
                        "application/x-bzip2",
                        "application/x-gzip",
                    ]:
                        # Read and rewind to ensure we're inside any compressed file layers
                        curfile.read(1)
                        curfile.seek(0)
                    else:
                        stop = True
                else:
                    stop = True

        if not IMPORTED_MAGIC:
            # Somewhat of a hack, but prevents a hard dependency on the magic module
            parsed_url = urllib.parse.urlparse(url)
            url_path = parsed_url.path
            stop = False
            while not stop:
                url_path_split = url_path.split(".")
                url_path_list, extension = url_path_split[:-1], url_path_split[-1]
                url_path = ".".join(url_path_list)
                if extension == "xz":
                    curfile = cascadeCloseFile(lzma.LZMAFile(curfile, mode), curfile)
                elif extension == "bz2":
                    curfile = cascadeCloseFile(bz2.BZ2File(curfile, mode), curfile)
                elif extension == "gz":
                    curfile = cascadeCloseFile(
                        gzip.GzipFile(fileobj=curfile, mode=mode), curfile
                    )
                else:
                    stop = True

        # Fallback in case the file doesn't exist
        if curfile is None:
            raise ValueError("URL does not reference an openable file")
        return curfile


class VolatilityHandler(urllib.request.BaseHandler):
    @classmethod
    def non_cached_schemes(cls) -> List[str]:
        return []


class JarHandler(VolatilityHandler):
    """Handles the jar scheme for URIs.

    Reference used for the schema syntax:
    http://docs.netkernel.org/book/view/book:mod:reference/doc:layer1:schemes:jar

    Actual reference (found from https://www.w3.org/wiki/UriSchemes/jar) seemed not to return:
    http://developer.java.sun.com/developer/onlineTraining/protocolhandlers/
    """

    @classmethod
    def non_cached_schemes(cls) -> List[str]:
        return ["jar"]

    @staticmethod
    def default_open(req: urllib.request.Request) -> Optional[Any]:
        """Handles the request if it's the jar scheme."""
        if req.type == "jar":
            subscheme, remainder = (
                req.full_url.split(":")[1],
                ":".join(req.full_url.split(":")[2:]),
            )
            if subscheme != "file":
                vollog.log(
                    constants.LOGLEVEL_VVV, f"Unsupported jar subscheme {subscheme}"
                )
                return None

            zipsplit = remainder.split("!")
            if len(zipsplit) != 2:
                vollog.log(
                    constants.LOGLEVEL_VVV,
                    f"Path did not contain exactly one fragment indicator: {remainder}",
                )
                return None

            zippath, filepath = zipsplit
            return zipfile.ZipFile(zippath).open(filepath)
        return None


class OfflineHandler(VolatilityHandler):
    @staticmethod
    def default_open(req: urllib.request.Request) -> Optional[Any]:
        if constants.OFFLINE and req.type in ["http", "https"]:
            raise exceptions.OfflineException(req.full_url)
        return None



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Optional, Tuple

from volatility3.framework import constants, interfaces
from volatility3.framework.automagic import symbol_cache, symbol_finder
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import intel, scanners
from volatility3.framework.symbols import linux

vollog = logging.getLogger(__name__)


class LinuxIntelStacker(interfaces.automagic.StackerLayerInterface):
    stack_order = 35
    exclusion_list = ["mac", "windows"]

    @classmethod
    def stack(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[interfaces.layers.DataLayerInterface]:
        """Attempts to identify linux within this layer."""
        # Bail out by default unless we can stack properly
        layer = context.layers[layer_name]
        join = interfaces.configuration.path_join

        # Never stack on top of an intel layer
        # FIXME: Find a way to improve this check
        if isinstance(layer, intel.Intel):
            return None

        linux_banners = symbol_cache.load_cache_manager().get_identifier_dictionary(
            operating_system="linux"
        )
        # If we have no banners, don't bother scanning
        if not linux_banners:
            vollog.info(
                "No Linux banners found - if this is a linux plugin, please check your symbol files location"
            )
            return None

        mss = scanners.MultiStringScanner([x for x in linux_banners if x is not None])
        for _, banner in layer.scan(
            context=context, scanner=mss, progress_callback=progress_callback
        ):
            dtb = None
            vollog.debug(f"Identified banner: {repr(banner)}")

            isf_path = linux_banners.get(banner, None)
            if isf_path:
                table_name = context.symbol_space.free_table_name("LintelStacker")
                table = linux.LinuxKernelIntermedSymbols(
                    context,
                    "temporary." + table_name,
                    name=table_name,
                    isf_url=isf_path,
                )
                context.symbol_space.append(table)

                kaslr_shift, aslr_shift = cls.find_aslr(
                    context,
                    table_name,
                    layer_name,
                    progress_callback=progress_callback,
                )

                if "init_top_pgt" in table.symbols:
                    layer_class = intel.LinuxIntel32e
                    dtb_symbol_name = "init_top_pgt"
                elif "init_level4_pgt" in table.symbols:
                    layer_class = intel.LinuxIntel32e
                    dtb_symbol_name = "init_level4_pgt"
                elif "pkmap_count" in table.symbols and table.get_symbol(
                    "pkmap_count"
                ).type.count in (512, 2048):
                    layer_class = intel.LinuxIntelPAE
                    dtb_symbol_name = "swapper_pg_dir"
                else:
                    layer_class = intel.LinuxIntel
                    dtb_symbol_name = "swapper_pg_dir"

                dtb = cls.virtual_to_physical_address(
                    table.get_symbol(dtb_symbol_name).address + kaslr_shift
                )

                # Build the new layer
                new_layer_name = context.layers.free_layer_name("IntelLayer")
                config_path = join("IntelHelper", new_layer_name)
                context.config[join(config_path, "memory_layer")] = layer_name
                context.config[join(config_path, "page_map_offset")] = dtb
                context.config[
                    join(config_path, LinuxSymbolFinder.banner_config_key)
                ] = str(banner, "latin-1")

                layer = layer_class(
                    context,
                    config_path=config_path,
                    name=new_layer_name,
                    metadata={"os": "Linux"},
                )
                layer.config["kernel_virtual_offset"] = aslr_shift

            if layer and dtb:
                vollog.debug(f"DTB was found at: 0x{dtb:0x}")
                return layer
        vollog.debug("No suitable linux banner could be matched")
        return None

    @classmethod
    def find_aslr(
        cls,
        context: interfaces.context.ContextInterface,
        symbol_table: str,
        layer_name: str,
        progress_callback: constants.ProgressCallback = None,
    ) -> Tuple[int, int]:
        """Determines the offset of the actual DTB in physical space and its
        symbol offset.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            symbol_table: The name of the kernel module on which to operate
            layer_name: The layer within the context in which the module exists
            progress_callback: A function that takes a percentage (and an optional description) that will be called periodically

        Returns:
            kaslr_shirt and aslr_shift
        """
        init_task_symbol = symbol_table + constants.BANG + "init_task"
        init_task_json_address = context.symbol_space.get_symbol(
            init_task_symbol
        ).address
        swapper_signature = rb"swapper(\/0|\x00\x00)\x00\x00\x00\x00\x00\x00"
        module = context.module(symbol_table, layer_name, 0)
        address_mask = context.symbol_space[symbol_table].config.get(
            "symbol_mask", None
        )

        task_symbol = module.get_type("task_struct")
        comm_child_offset = task_symbol.relative_child_offset("comm")

        for offset in context.layers[layer_name].scan(
            scanner=scanners.RegExScanner(swapper_signature),
            context=context,
            progress_callback=progress_callback,
        ):
            init_task_address = offset - comm_child_offset
            init_task = module.object(
                object_type="task_struct", offset=init_task_address, absolute=True
            )
            if init_task.pid != 0:
                continue
            elif (
                init_task.has_member("state")
                and init_task.state.cast("unsigned int") != 0
            ):
                continue
            elif init_task.active_mm.cast("long unsigned int") == module.get_symbol(
                "init_mm"
            ).address and init_task.tasks.next.cast(
                "long unsigned int"
            ) == init_task.tasks.prev.cast("long unsigned int"):
                # The idle task steals `mm` from previously running task, i.e.,
                # `init_mm` is only used as long as no CPU has ever been idle.
                # This catches cases where we found a fragment of the
                # unrelocated ELF file instead of the running kernel.
                continue

            # This we get for free
            aslr_shift = (
                int.from_bytes(
                    init_task.files.cast("bytes", length=init_task.files.vol.size),
                    byteorder=init_task.files.vol.data_format.byteorder,
                )
                - module.get_symbol("init_files").address
            )
            kaslr_shift = init_task_address - cls.virtual_to_physical_address(
                init_task_json_address
            )
            if address_mask:
                aslr_shift = aslr_shift & address_mask

            if aslr_shift & 0xFFF != 0 or kaslr_shift & 0xFFF != 0:
                continue
            vollog.debug(
                f"Linux ASLR shift values determined: physical {kaslr_shift:0x} virtual {aslr_shift:0x}"
            )
            return kaslr_shift, aslr_shift

        # We don't throw an exception, because we may legitimately not have an ASLR shift, but we report it
        vollog.debug("Scanners could not determine any ASLR shifts, using 0 for both")
        return 0, 0

    @staticmethod
    def virtual_to_physical_address(addr: int) -> int:
        """Converts a virtual linux address to a physical one (does not account
        of ASLR)"""
        if addr > 0xFFFFFFFF80000000:
            return addr - 0xFFFFFFFF80000000
        return addr - 0xC0000000


class LinuxSymbolFinder(symbol_finder.SymbolFinder):
    """Linux symbol loader based on uname signature strings."""

    banner_config_key = "kernel_banner"
    operating_system = "linux"
    symbol_class = "volatility3.framework.symbols.linux.LinuxKernelIntermedSymbols"
    exclusion_list = ["mac", "windows"]

    @classmethod
    def find_aslr(cls, *args):
        return LinuxIntelStacker.find_aslr(*args)[1]


class LinuxIntelVMCOREINFOStacker(interfaces.automagic.StackerLayerInterface):
    stack_order = 34
    exclusion_list = ["mac", "windows"]

    @staticmethod
    def _check_versions() -> bool:
        """Verify the versions of the required modules"""
        # Check VMCOREINFO API version
        vmcoreinfo_version_required = (1, 0, 0)
        if not requirements.VersionRequirement.matches_required(
            vmcoreinfo_version_required, linux.VMCoreInfo.version
        ):
            vollog.info(
                "VMCOREINFO version not suitable: required %s found %s",
                vmcoreinfo_version_required,
                linux.VMCoreInfo.version,
            )
            return False

        return True

    @classmethod
    def stack(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[interfaces.layers.DataLayerInterface]:
        """Attempts to identify linux within this layer."""

        # Verify the versions of the required modules
        if not cls._check_versions():
            return None

        # Bail out by default unless we can stack properly
        layer = context.layers[layer_name]

        # Never stack on top of an intel layer
        # FIXME: Find a way to improve this check
        if isinstance(layer, intel.Intel):
            return None

        linux_banners = symbol_cache.load_cache_manager().get_identifier_dictionary(
            operating_system="linux"
        )
        if not linux_banners:
            # If we have no banners, don't bother scanning
            vollog.info(
                "No Linux banners found - if this is a linux plugin, please check your "
                "symbol files location"
            )
            return None

        vmcoreinfo_elf_notes_iter = linux.VMCoreInfo.search_vmcoreinfo_elf_note(
            context=context,
            layer_name=layer_name,
            progress_callback=progress_callback,
        )

        # Iterate through each VMCOREINFO ELF note found, using the first one that is valid.
        for _vmcoreinfo_offset, vmcoreinfo in vmcoreinfo_elf_notes_iter:
            shifts = cls._vmcoreinfo_find_aslr(vmcoreinfo)
            if not shifts:
                # Let's try the next VMCOREINFO, in case this one isn't correct.
                continue

            kaslr_shift, aslr_shift = shifts

            dtb = cls._vmcoreinfo_get_dtb(vmcoreinfo, aslr_shift, kaslr_shift)
            if dtb is None:
                # Discard this VMCOREINFO immediately
                continue

            is_32bit, is_pae = cls._vmcoreinfo_is_32bit(vmcoreinfo)
            if is_32bit:
                layer_class = intel.IntelPAE if is_pae else intel.Intel
            else:
                layer_class = intel.Intel32e

            uts_release = vmcoreinfo["OSRELEASE"]

            # See how linux_banner constant is built in the linux kernel
            linux_version_prefix = f"Linux version {uts_release} (".encode()
            valid_banners = [
                x for x in linux_banners if x and x.startswith(linux_version_prefix)
            ]
            if not valid_banners:
                # There's no banner matching this VMCOREINFO, keep trying with the next one
                continue
            elif len(valid_banners) == 1:
                # Usually, we narrow down the Linux banner list to a single element.
                # Using BytesScanner here is slightly faster than MultiStringScanner.
                scanner = scanners.BytesScanner(valid_banners[0])
            else:
                scanner = scanners.MultiStringScanner(valid_banners)

            join = interfaces.configuration.path_join
            for match in layer.scan(
                context=context, scanner=scanner, progress_callback=progress_callback
            ):
                # Unfortunately, the scanners do not maintain a consistent interface
                banner = match[1] if isinstance(match, Tuple) else valid_banners[0]

                isf_path = linux_banners.get(banner, None)
                if not isf_path:
                    vollog.warning(
                        "Identified banner %r, but no matching ISF is available.",
                        banner,
                    )
                    continue

                vollog.debug("Identified banner: %r", banner)
                table_name = context.symbol_space.free_table_name("LintelStacker")
                table = linux.LinuxKernelIntermedSymbols(
                    context,
                    f"temporary.{table_name}",
                    name=table_name,
                    isf_url=isf_path,
                )
                context.symbol_space.append(table)

                # Build the new layer
                new_layer_name = context.layers.free_layer_name("primary")
                config_path = join("vmcoreinfo", new_layer_name)
                kernel_banner = LinuxSymbolFinder.banner_config_key
                banner_str = banner.decode(encoding="latin-1")
                context.config[join(config_path, kernel_banner)] = banner_str
                context.config[join(config_path, "memory_layer")] = layer_name
                context.config[join(config_path, "page_map_offset")] = dtb
                context.config[join(config_path, "kernel_virtual_offset")] = aslr_shift
                layer = layer_class(
                    context,
                    config_path=config_path,
                    name=new_layer_name,
                    metadata={"os": "Linux"},
                )

                if layer:
                    vollog.debug(
                        "Values found in VMCOREINFO: KASLR=0x%x, ASLR=0x%x, DTB=0x%x",
                        kaslr_shift,
                        aslr_shift,
                        dtb,
                    )

                    return layer

        vollog.debug("No suitable linux banner could be matched")
        return None

    @staticmethod
    def _vmcoreinfo_find_aslr(vmcoreinfo) -> Tuple[int, int]:
        phys_base = vmcoreinfo.get("NUMBER(phys_base)")
        if phys_base is None:
            # In kernel < 4.10, there may be a SYMBOL(phys_base), but as noted in the
            # c401721ecd1dcb0a428aa5d6832ee05ffbdbffbbe commit comment, this value
            # isn't useful for calculating the physical address.
            # There's nothing we can do here, so let's try with the next VMCOREINFO or
            # the next Stacker.
            return None

        # kernels 3.14 (b6085a865762236bb84934161273cdac6dd11c2d) KERNELOFFSET was added
        kerneloffset = vmcoreinfo.get("KERNELOFFSET")
        if kerneloffset is None:
            # kernels < 3.14 if KERNELOFFSET is missing, KASLR might not be implemented.
            # Oddly, NUMBER(phys_base) is present without it. To be safe, proceed only
            # if both are present.
            return None

        aslr_shift = kerneloffset
        kaslr_shift = phys_base + aslr_shift

        return kaslr_shift, aslr_shift

    @staticmethod
    def _vmcoreinfo_get_dtb(vmcoreinfo, aslr_shift, kaslr_shift) -> int:
        """Returns the page global directory physical address (a.k.a DTB or PGD)"""
        # In x86-64, since kernels 2.5.22 swapper_pg_dir is a macro to the respective pgd.
        # First, in e3ebadd95cb621e2c7436f3d3646447ac9d5c16d to init_level4_pgt, and later
        # in kernels 4.13 in 65ade2f872b474fa8a04c2d397783350326634e6) to init_top_pgt.
        # In x86-32, the pgd is swapper_pg_dir. So, in any case, for VMCOREINFO
        # SYMBOL(swapper_pg_dir) will always have the right value.
        dtb_vaddr = vmcoreinfo.get("SYMBOL(swapper_pg_dir)")
        if dtb_vaddr is None:
            # Abort, it should be present
            return None

        dtb_paddr = (
            LinuxIntelStacker.virtual_to_physical_address(dtb_vaddr)
            - aslr_shift
            + kaslr_shift
        )

        return dtb_paddr

    @staticmethod
    def _vmcoreinfo_is_32bit(vmcoreinfo) -> Tuple[bool, bool]:
        """Returns a tuple of booleans with is_32bit and is_pae values"""
        is_pae = vmcoreinfo.get("CONFIG_X86_PAE", "n") == "y"
        if is_pae:
            is_32bit = True
        else:
            # Check the swapper_pg_dir virtual address size
            dtb_vaddr = vmcoreinfo["SYMBOL(swapper_pg_dir)"]
            is_32bit = dtb_vaddr <= 2**32

        return is_32bit, is_pae



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""Volatility 3 Linux Constants.

Linux-specific values that aren't found in debug symbols
"""

import enum
from dataclasses import dataclass

# Exec argument limits
# Ref: include/uapi/linux/binfmts.h (linux.git commit f6031913338f1dad5bd8cb7286ff4e53644b6940)
MAX_ARG_STRLEN = 32 * 4096

KERNEL_NAME = "__kernel__"

"""The value hard coded from the Linux Kernel (hence not extracted from the layer itself)"""

# include/linux/sched.h
PF_KTHREAD = 0x00200000  # I'm a kernel thread

# Standard well-defined IP protocols.
# ref: include/uapi/linux/in.h
IP_PROTOCOLS = {
    0: "IP",
    1: "ICMP",
    2: "IGMP",
    4: "IPIP",
    6: "TCP",
    8: "EGP",
    12: "PUP",
    17: "UDP",
    22: "IDP",
    29: "TP",
    33: "DCCP",
    41: "IPV6",
    46: "RSVP",
    47: "GRE",
    50: "ESP",
    51: "AH",
    92: "MTP",
    94: "BEETPH",
    98: "ENCAP",
    103: "PIM",
    108: "COMP",
    132: "SCTP",
    136: "UDPLITE",
    137: "MPLS",
    143: "ETHERNET",
    255: "RAW",
    262: "MPTCP",
}

# IPV6 extension headers
# ref: include/uapi/linux/in6.h
IPV6_PROTOCOLS = {
    0: "HOPBYHOP_OPTS",
    43: "ROUTING",
    44: "FRAGMENT",
    58: "ICMPv6",
    59: "NO_NEXT",
    60: "DESTINATION_OPTS",
    135: "MOBILITY",
}

# ref: include/net/tcp_states.h
TCP_STATES = (
    "",
    "ESTABLISHED",
    "SYN_SENT",
    "SYN_RECV",
    "FIN_WAIT1",
    "FIN_WAIT2",
    "TIME_WAIT",
    "CLOSE",
    "CLOSE_WAIT",
    "LAST_ACK",
    "LISTEN",
    "CLOSING",
    "TCP_NEW_SYN_RECV",
)

# ref: include/linux/net.h (socket_type enum)
SOCK_TYPES = {
    1: "STREAM",
    2: "DGRAM",
    3: "RAW",
    4: "RDM",
    5: "SEQPACKET",
    6: "DCCP",
    10: "PACKET",
}

# Address families
# ref: include/linux/socket.h
SOCK_FAMILY = (
    "AF_UNSPEC",
    "AF_UNIX",
    "AF_INET",
    "AF_AX25",
    "AF_IPX",
    "AF_APPLETALK",
    "AF_NETROM",
    "AF_BRIDGE",
    "AF_ATMPVC",
    "AF_X25",
    "AF_INET6",
    "AF_ROSE",
    "AF_DECnet",
    "AF_NETBEUI",
    "AF_SECURITY",
    "AF_KEY",
    "AF_NETLINK",
    "AF_PACKET",
    "AF_ASH",
    "AF_ECONET",
    "AF_ATMSVC",
    "AF_RDS",
    "AF_SNA",
    "AF_IRDA",
    "AF_PPPOX",
    "AF_WANPIPE",
    "AF_LLC",
    "AF_IB",
    "AF_MPLS",
    "AF_CAN",
    "AF_TIPC",
    "AF_BLUETOOTH",
    "AF_IUCV",
    "AF_RXRPC",
    "AF_ISDN",
    "AF_PHONET",
    "AF_IEEE802154",
    "AF_CAIF",
    "AF_ALG",
    "AF_NFC",
    "AF_VSOCK",
    "AF_KCM",
    "AF_QIPCRTR",
    "AF_SMC",
    "AF_XDP",
)

# Socket states
# ref: include/uapi/linux/net.h
SOCKET_STATES = ("FREE", "UNCONNECTED", "CONNECTING", "CONNECTED", "DISCONNECTING")

# Netlink protocols
# ref: include/uapi/linux/netlink.h
NETLINK_PROTOCOLS = (
    "NETLINK_ROUTE",
    "NETLINK_UNUSED",
    "NETLINK_USERSOCK",
    "NETLINK_FIREWALL",
    "NETLINK_SOCK_DIAG",
    "NETLINK_NFLOG",
    "NETLINK_XFRM",
    "NETLINK_SELINUX",
    "NETLINK_ISCSI",
    "NETLINK_AUDIT",
    "NETLINK_FIB_LOOKUP",
    "NETLINK_CONNECTOR",
    "NETLINK_NETFILTER",
    "NETLINK_IP6_FW",
    "NETLINK_DNRTMSG",
    "NETLINK_KOBJECT_UEVENT",
    "NETLINK_GENERIC",
    "NETLINK_DM",
    "NETLINK_SCSITRANSPORT",
    "NETLINK_ECRYPTFS",
    "NETLINK_RDMA",
    "NETLINK_CRYPTO",
    "NETLINK_SMC",
)

# Short list of Ethernet Protocol ID's.
# ref: include/uapi/linux/if_ether.h
# Used in AF_PACKET socket family
ETH_PROTOCOLS = {
    0x0001: "ETH_P_802_3",
    0x0002: "ETH_P_AX25",
    0x0003: "ETH_P_ALL",
    0x0004: "ETH_P_802_2",
    0x0005: "ETH_P_SNAP",
    0x0006: "ETH_P_DDCMP",
    0x0007: "ETH_P_WAN_PPP",
    0x0008: "ETH_P_PPP_MP",
    0x0009: "ETH_P_LOCALTALK",
    0x000C: "ETH_P_CAN",
    0x000F: "ETH_P_CANFD",
    0x0010: "ETH_P_PPPTALK",
    0x0011: "ETH_P_TR_802_2",
    0x0016: "ETH_P_CONTROL",
    0x0017: "ETH_P_IRDA",
    0x0018: "ETH_P_ECONET",
    0x0019: "ETH_P_HDLC",
    0x001A: "ETH_P_ARCNET",
    0x001B: "ETH_P_DSA",
    0x001C: "ETH_P_TRAILER",
    0x0060: "ETH_P_LOOP",
    0x00F6: "ETH_P_IEEE802154",
    0x00F7: "ETH_P_CAIF",
    0x00F8: "ETH_P_XDSA",
    0x00F9: "ETH_P_MAP",
    0x0800: "ETH_P_IP",
    0x0805: "ETH_P_X25",
    0x0806: "ETH_P_ARP",
    0x8035: "ETH_P_RARP",
    0x809B: "ETH_P_ATALK",
    0x80F3: "ETH_P_AARP",
    0x8100: "ETH_P_8021Q",
}

# Connection and socket states
# ref: include/net/bluetooth/bluetooth.h
BLUETOOTH_STATES = (
    "",
    "CONNECTED",
    "OPEN",
    "BOUND",
    "LISTEN",
    "CONNECT",
    "CONNECT2",
    "CONFIG",
    "DISCONN",
    "CLOSED",
)

# Bluetooth protocols
# ref: include/net/bluetooth/bluetooth.h
BLUETOOTH_PROTOCOLS = (
    "L2CAP",
    "HCI",
    "SCO",
    "RFCOMM",
    "BNEP",
    "CMTP",
    "HIDP",
    "AVDTP",
)

# Ref: include/uapi/linux/capability.h
CAPABILITIES = (
    "chown",
    "dac_override",
    "dac_read_search",
    "fowner",
    "fsetid",
    "kill",
    "setgid",
    "setuid",
    "setpcap",
    "linux_immutable",
    "net_bind_service",
    "net_broadcast",
    "net_admin",
    "net_raw",
    "ipc_lock",
    "ipc_owner",
    "sys_module",
    "sys_rawio",
    "sys_chroot",
    "sys_ptrace",
    "sys_pacct",
    "sys_admin",
    "sys_boot",
    "sys_nice",
    "sys_resource",
    "sys_time",
    "sys_tty_config",
    "mknod",
    "lease",
    "audit_write",
    "audit_control",
    "setfcap",
    "mac_override",
    "mac_admin",
    "syslog",
    "wake_alarm",
    "block_suspend",
    "audit_read",
    "perfmon",
    "bpf",
    "checkpoint_restore",
)

ELF_MAX_EXTRACTION_SIZE = 1024 * 1024 * 1024 * 4 - 1

# For IFA_* below - Ref: include/net/ipv6.h
IPV6_ADDR_LOOPBACK = 0x0010
IPV6_ADDR_LINKLOCAL = 0x0020
IPV6_ADDR_SITELOCAL = 0x0040
# For inet6_ifaddr - Ref: include/net/if_inet6.h
IFA_HOST = IPV6_ADDR_LOOPBACK
IFA_LINK = IPV6_ADDR_LINKLOCAL
IFA_SITE = IPV6_ADDR_SITELOCAL

# Only for kernels < 3.15 when the net_device_flags enum didn't exist
# ref include/uapi/linux/if.h
NET_DEVICE_FLAGS = {
    "IFF_UP": 0x1,
    "IFF_BROADCAST": 0x2,
    "IFF_DEBUG": 0x4,
    "IFF_LOOPBACK": 0x8,
    "IFF_POINTOPOINT": 0x10,
    "IFF_NOTRAILERS": 0x20,
    "IFF_RUNNING": 0x40,
    "IFF_NOARP": 0x80,
    "IFF_PROMISC": 0x100,
    "IFF_ALLMULTI": 0x200,
    "IFF_MASTER": 0x400,
    "IFF_SLAVE": 0x800,
    "IFF_MULTICAST": 0x1000,
    "IFF_PORTSEL": 0x2000,
    "IFF_AUTOMEDIA": 0x4000,
    "IFF_DYNAMIC": 0x8000,
    "IFF_LOWER_UP": 0x10000,
    "IFF_DORMANT": 0x20000,
    "IFF_ECHO": 0x40000,
}


# Kernels >= 2.6.17. See IF_OPER_* in include/uapi/linux/if.h
class IF_OPER_STATES(enum.Enum):
    """RFC 2863 - Network interface operational status"""

    UNKNOWN = 0
    NOTPRESENT = 1
    DOWN = 2
    LOWERLAYERDOWN = 3
    TESTING = 4
    DORMANT = 5
    UP = 6


class ELF_IDENT(enum.IntEnum):
    """ELF header e_ident indexes"""

    EI_MAG0 = 0
    EI_MAG1 = 1
    EI_MAG2 = 2
    EI_MAG3 = 3
    EI_CLASS = 4
    EI_DATA = 5
    EI_VERSION = 6
    EI_OSABI = 7
    EI_PAD = 8


class ELF_CLASS(enum.IntEnum):
    """ELF header class types"""

    ELFCLASSNONE = 0
    ELFCLASS32 = 1
    ELFCLASS64 = 2


# PTrace
PT_OPT_FLAG_SHIFT = 3

PTRACE_EVENT_FORK = 1
PTRACE_EVENT_VFORK = 2
PTRACE_EVENT_CLONE = 3
PTRACE_EVENT_EXEC = 4
PTRACE_EVENT_VFORK_DONE = 5
PTRACE_EVENT_EXIT = 6
PTRACE_EVENT_SECCOMP = 7

PTRACE_O_EXITKILL = 1 << 20
PTRACE_O_SUSPEND_SECCOMP = 1 << 21


class PT_FLAGS(enum.Flag):
    "PTrace flags"

    PT_PTRACED = 0x00001
    PT_SEIZED = 0x10000

    PT_TRACESYSGOOD = 1 << (PT_OPT_FLAG_SHIFT + 0)
    PT_TRACE_FORK = 1 << (PT_OPT_FLAG_SHIFT + PTRACE_EVENT_FORK)
    PT_TRACE_VFORK = 1 << (PT_OPT_FLAG_SHIFT + PTRACE_EVENT_VFORK)
    PT_TRACE_CLONE = 1 << (PT_OPT_FLAG_SHIFT + PTRACE_EVENT_CLONE)
    PT_TRACE_EXEC = 1 << (PT_OPT_FLAG_SHIFT + PTRACE_EVENT_EXEC)
    PT_TRACE_VFORK_DONE = 1 << (PT_OPT_FLAG_SHIFT + PTRACE_EVENT_VFORK_DONE)
    PT_TRACE_EXIT = 1 << (PT_OPT_FLAG_SHIFT + PTRACE_EVENT_EXIT)
    PT_TRACE_SECCOMP = 1 << (PT_OPT_FLAG_SHIFT + PTRACE_EVENT_SECCOMP)

    PT_EXITKILL = PTRACE_O_EXITKILL << PT_OPT_FLAG_SHIFT
    PT_SUSPEND_SECCOMP = PTRACE_O_SUSPEND_SECCOMP << PT_OPT_FLAG_SHIFT

    @property
    def flags(self) -> str:
        """Returns the ptrace flags string"""
        return str(self).replace(self.__class__.__name__ + ".", "")


# Boot time
NSEC_PER_SEC = 1e9


# Valid sizes for modules. Note that the Linux kernel does not define these values; they
# are based on empirical observations of typical memory allocations for kernel modules.
# We use this to verify that the found module falls within reasonable limits.
MODULE_MAXIMUM_CORE_SIZE = 20000000
MODULE_MAXIMUM_CORE_TEXT_SIZE = 20000000
MODULE_MINIMUM_SIZE = 4096

# Kallsyms
KSYM_NAME_LEN = 512
NM_TYPES_DESC = {
    "a": "Symbol is absolute and doesn't change during linking",
    "b": "Symbol in the BSS section, typically holding zero-initialized or uninitialized data",
    "c": "Symbol is common, typically holding uninitialized data",
    "d": "Symbol is in the initialized data section",
    "g": "Symbol is in an initialized data section for small objects",
    "i": "Symbol is an indirect reference to another symbol",
    "N": "Symbol is a debugging symbol",
    "n": "Symbol is in a non-data, non-code, non-debug read-only section",
    "p": "Symbol is in a stack unwind section",
    "r": "Symbol is in a read only data section",
    "s": "Symbol is in an uninitialized or zero-initialized data section for small objects",
    "t": "Symbol is in the text (code) section",
    "U": "Symbol is undefined",
    "u": "Symbol is a unique global symbol",
    "V": "Symbol is a weak object, with a default value",
    "v": "Symbol is a weak object",
    "W": "Symbol is a weak symbol but not marked as a weak object symbol, with a default value",
    "w": "Symbol is a weak symbol but not marked as a weak object symbol",
    "?": "Symbol type is unknown",
}

# VMCOREINFO
VMCOREINFO_MAGIC = b"VMCOREINFO\x00"
# Aligned to 4 bytes. See storenote() in kernels < 4.19 or append_kcore_note() in kernels >= 4.19
VMCOREINFO_MAGIC_ALIGNED = VMCOREINFO_MAGIC + b"\x00"
OSRELEASE_TAG = b"OSRELEASE="

ATTRIBUTE_NAME_MAX_SIZE = 255
"""
In 5.9-rc1+, the Linux kernel limits the READ size of a section bin_attribute name to MODULE_SECT_READ_SIZE:

- https://elixir.bootlin.com/linux/v6.15-rc4/source/kernel/module/sysfs.c#L106
- https://github.com/torvalds/linux/commit/11990a5bd7e558e9203c1070fc52fb6f0488e75b

However, the raw section name loaded from the .ko ELF can in theory be thousands of characters,
and unless we do a NULL terminated search we can't set a perfect value.
"""


@dataclass
class TaintFlag:
    shift: int
    desc: str
    when_present: bool
    module: bool


TAINT_FLAGS = {
    "P": TaintFlag(
        shift=1 << 0, desc="PROPRIETARY_MODULE", when_present=True, module=True
    ),
    "G": TaintFlag(
        shift=1 << 0, desc="PROPRIETARY_MODULE", when_present=False, module=True
    ),
    "F": TaintFlag(shift=1 << 1, desc="FORCED_MODULE", when_present=True, module=False),
    "S": TaintFlag(
        shift=1 << 2, desc="CPU_OUT_OF_SPEC", when_present=True, module=False
    ),
    "R": TaintFlag(shift=1 << 3, desc="FORCED_RMMOD", when_present=True, module=False),
    "M": TaintFlag(shift=1 << 4, desc="MACHINE_CHECK", when_present=True, module=False),
    "B": TaintFlag(shift=1 << 5, desc="BAD_PAGE", when_present=True, module=False),
    "U": TaintFlag(shift=1 << 6, desc="USER", when_present=True, module=False),
    "D": TaintFlag(shift=1 << 7, desc="DIE", when_present=True, module=False),
    "A": TaintFlag(
        shift=1 << 8, desc="OVERRIDDEN_ACPI_TABLE", when_present=True, module=False
    ),
    "W": TaintFlag(shift=1 << 9, desc="WARN", when_present=True, module=False),
    "C": TaintFlag(shift=1 << 10, desc="CRAP", when_present=True, module=True),
    "I": TaintFlag(
        shift=1 << 11, desc="FIRMWARE_WORKAROUND", when_present=True, module=False
    ),
    "O": TaintFlag(shift=1 << 12, desc="OOT_MODULE", when_present=True, module=True),
    "E": TaintFlag(
        shift=1 << 13, desc="UNSIGNED_MODULE", when_present=True, module=True
    ),
    "L": TaintFlag(shift=1 << 14, desc="SOFTLOCKUP", when_present=True, module=False),
    "K": TaintFlag(shift=1 << 15, desc="LIVEPATCH", when_present=True, module=True),
    "X": TaintFlag(shift=1 << 16, desc="AUX", when_present=True, module=True),
    "T": TaintFlag(shift=1 << 17, desc="RANDSTRUCT", when_present=True, module=True),
    "N": TaintFlag(shift=1 << 18, desc="TEST", when_present=True, module=True),
}
"""Flags used to taint kernel and modules, for debugging purposes.

Map based on 6.12-rc5.

Documentation :
    - https://www.kernel.org/doc/Documentation/admin-guide/sysctl/kernel.rst#:~:text=guide/sysrq.rst.-,tainted,-%3D%3D%3D%3D%3D%3D%3D%0A%0ANon%2Dzero%20if
    - https://www.kernel.org/doc/Documentation/admin-guide/tainted-kernels.rst#:~:text=More%20detailed%20explanation%20for%20tainting
    - taint_flag kernel struct
    - taint_flags kernel constant
"""

## ELF related constants

# Elf Symbol Bindings
STB_LOCAL = 0
STB_GLOBAL = 1

# Elf Symbol Types
STT_NOTYPE = 0
STT_OBJECT = 1
STT_FUNC = 2
STT_SECTION = 3

# Elf Section Types
SHT_NULL = 0
SHT_PROGBITS = 1
SHT_SYMTAB = 2
SHT_STRTAB = 3
SHT_RELA = 4
SHT_NOTE = 7

# Elf Section Attributes
SHF_WRITE = 1
SHF_ALLOC = 2
SHF_EXECINSTR = 4



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Generator, Iterable, List, Tuple

from volatility3.framework import constants, interfaces, objects, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import configuration
from volatility3.framework.objects.utility import array_to_string
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows.extensions import pe
from volatility3.plugins.windows import modules

vollog = logging.getLogger(__name__)


class Passphrase(interfaces.plugins.PluginInterface):
    """TrueCrypt Cached Passphrase Finder"""

    _version = (0, 1, 0)
    _required_framework_version = (2, 5, 2)

    @classmethod
    def get_requirements(cls) -> List[configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                "kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="modules", component=modules.Modules, version=(3, 0, 0)
            ),
            requirements.IntRequirement(
                name="min-length",
                description="Minimum length of passphrases to identify",
                default=5,
                optional=True,
            ),
        ]

    def scan_module(
        self, module_base: int, layer_name: str
    ) -> Generator[Tuple[int, str], None, None]:
        """Scans the TrueCrypt kernel module for cached passphrases.

        Args:
            module_base: the module's DLL base
            layer_name: the name of the layer in which the module resides

        Generates:
            A tuple of the offset at which a password is found, and the password
        """
        pe_table_name = intermed.IntermediateSymbolTable.create(
            self.context, self.config_path, "windows", "pe", class_types=pe.class_types
        )
        dos_header: pe.IMAGE_DOS_HEADER = self.context.object(
            pe_table_name + constants.BANG + "_IMAGE_DOS_HEADER",
            layer_name,
            module_base,
        )
        data_section: objects.StructType = next(
            sec
            for sec in dos_header.get_nt_header().get_sections()
            if array_to_string(sec.Name) == ".data"
        )
        base: int = data_section.VirtualAddress + module_base
        size: int = data_section.Misc.VirtualSize
        # Looking at `Length` in TrueCrypt/Common/Password.h::Password struct
        DWORD_SIZE_BYTES: int = 4
        format = objects.DataFormatInfo(
            length=DWORD_SIZE_BYTES, byteorder="little", signed=True
        )
        int32 = objects.templates.ObjectTemplate(
            objects.Integer, pe_table_name + constants.BANG + "int", data_format=format
        )
        count, not_aligned = divmod(size, DWORD_SIZE_BYTES)
        if not_aligned:
            raise ValueError("PE data section not DWORD-aligned!")
        lengths = self.context.object(
            pe_table_name + constants.BANG + "array",
            layer_name,
            base,
            count=count,
            subtype=int32,
        )
        min_length = self.config.get("min-length")
        for length in lengths:
            # TrueCrypt maximum password length is 64
            # (see TrueCrypt/Common/Password.h)
            if not min_length <= length <= 64:
                continue
            offset = length.vol["offset"] + DWORD_SIZE_BYTES
            passphrase: objects.Bytes = self.context.object(
                pe_table_name + constants.BANG + "bytes",
                layer_name,
                offset,
                length=length,
            )
            # TrueCrypt/Common/Password.c permits chars in the range
            # [0x20, 0x7F).
            if not all(0x20 <= c < 0x7F for c in passphrase):
                continue
            # TrueCrypt/Common/Password.h::Password struct is padded with
            # 3 zero bytes to keep 64-byte alignment.
            buf: objects.Bytes = self.context.object(
                pe_table_name + constants.BANG + "bytes",
                layer_name,
                offset + length + 1,  # +1 for '\0'-terminated password string
                length=3,
            )
            if any(buf):
                continue
            # Password found.
            yield offset, passphrase.decode(encoding="ascii")

    def _generator(self):
        kernel = self.context.modules[self.config["kernel"]]
        mods: Iterable[interfaces.objects.ObjectInterface] = (
            modules.Modules.list_modules(self.context, self.config["kernel"])
        )
        try:
            truecrypt_module_base = next(
                mod.DllBase
                for mod in mods
                if mod.BaseDllName.get_string().lower() == "truecrypt.sys"
            )
        except StopIteration:
            vollog.warning(
                "Truecrypt module not found in the modules list. Unable to proceed."
            )
            return

        for offset, password in self.scan_module(
            truecrypt_module_base, kernel.layer_name
        ):
            yield (0, (format_hints.Hex(offset), len(password), password))

    def run(self) -> renderers.TreeGrid:
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("Length", int),
                ("Password", str),
            ],
            self._generator(),
        )



import logging
from typing import Dict, Generator, List, Optional, Union

from volatility3.framework import objects, exceptions, renderers, interfaces, constants
from volatility3.framework.objects import utility
from volatility3.framework.constants import linux as linux_constants
from volatility3.framework.symbols import wrappers
from volatility3.framework.symbols import linux
from volatility3.framework.renderers import conversion
import socket as socket_module


vollog = logging.getLogger(__name__)


class net(objects.StructType):
    def get_inode(self) -> int:
        """Get the namespace id for this network namespace.

        Raises:
            AttributeError: If it cannot find the network namespace id for the
             current kernel.

        Returns:
            int: the namespace id
        """
        if self.has_member("proc_inum"):
            # 3.8.13 <= kernel < 3.19.8
            return self.proc_inum
        elif self.has_member("ns") and self.ns.has_member("inum"):
            # kernel >= 3.19.8
            return self.ns.inum
        else:
            # kernel < 3.8.13
            raise AttributeError("Unable to find net_namespace inode")


class net_device(objects.StructType):
    def get_device_name(self) -> str:
        """Return the network device name

        Returns:
            str: The network device name
        """
        return utility.array_to_string(self.name)

    def _format_as_mac_address(self, hwaddr) -> str:
        return ":".join([f"{x:02x}" for x in hwaddr[: self.addr_len]])

    def get_mac_address(self) -> Optional[str]:
        """Get the MAC address of this network interface.

        Returns:
            str: the MAC address of this network interface.
        """
        if self.has_member("perm_addr"):
            null_mac_addr_bytes = b"\x00" * self.addr_len
            null_mac_addr = self._format_as_mac_address(null_mac_addr_bytes)
            mac_addr = self._format_as_mac_address(self.perm_addr)
            if mac_addr != null_mac_addr:
                return mac_addr

        parent_layer = self._context.layers[self.vol.layer_name]
        try:
            hwaddr = parent_layer.read(self.dev_addr, self.addr_len, pad=True)
        except exceptions.InvalidAddressException:
            vollog.debug(
                f"Unable to read network interface mac address from {self.dev_addr:#x}"
            )
            return None

        return self._format_as_mac_address(hwaddr)

    def _get_flag_choices(self) -> Dict[str, int]:
        """Return the net_device flags as a list of strings"""
        vmlinux = linux.LinuxUtilities.get_module_from_volobj_type(self._context, self)
        try:
            # kernels >= 3.15
            net_device_flags_enum = vmlinux.get_enumeration("net_device_flags")
            choices = net_device_flags_enum.choices
        except exceptions.SymbolError:
            # kernels < 3.15
            choices = linux_constants.NET_DEVICE_FLAGS

        return choices

    def _get_net_device_flag_value(
        self, name
    ) -> Union[int, interfaces.renderers.BaseAbsentValue]:
        """Return the net_device flag value based on the flag name"""
        return self._get_flag_choices().get(name, renderers.UnparsableValue())

    def _get_netdev_state_t(self):
        vmlinux = linux.LinuxUtilities.get_module_from_volobj_type(self._context, self)
        try:
            # At least from kernels 2.6.30
            return vmlinux.get_enumeration("netdev_state_t")
        except exceptions.SymbolError:
            raise exceptions.VolatilityException(
                "Unsupported kernel or wrong ISF. Cannot find 'netdev_state_t' enumeration"
            )

    def is_running(self) -> bool:
        """Test if the network device has been brought up
        Based on netif_running()

        Returns:
            bool: True if the device is UP
        """
        netdev_state_t_enum = self._get_netdev_state_t()

        # It should be safe. netdev_state_t::__LINK_STATE_START has been available since
        # at least kernels 2.6.30
        return (
            self.state & (1 << netdev_state_t_enum.choices["__LINK_STATE_START"]) != 0
        )

    def is_carrier_ok(self) -> bool:
        """Check if carrier is present on network device
        Based on netif_carrier_ok()

        Returns:
            bool: True if carrier present
        """
        netdev_state_t_enum = self._get_netdev_state_t()

        # It should be safe. netdev_state_t::__LINK_STATE_NOCARRIER has been available
        # since at least kernels 2.6.30
        return (
            self.state & (1 << netdev_state_t_enum.choices["__LINK_STATE_NOCARRIER"])
            == 0
        )

    def is_dormant(self) -> bool:
        """Check if the network device is dormant
        Based on netif_dormant(()

        Returns:
            bool: True if the network device is dormant
        """
        netdev_state_t_enum = self._get_netdev_state_t()

        # It should be safe. netdev_state_t::__LINK_STATE_DORMANT has been available
        # since at least kernels 2.6.30
        return (
            self.state & (1 << netdev_state_t_enum.choices["__LINK_STATE_DORMANT"]) != 0
        )

    def is_operational(self) -> bool:
        """Test if the carrier is operational
        Based on netif_oper_up()

        Returns:
            bool: True if the device is UP
        """

        return self.get_operational_state() in ("UP", "UNKNOWN")

    def get_flag_names(self) -> List[str]:
        """Return the net_device flags as a list of strings.
        This is the combination of flags exported through kernel APIs to userspace.
        Based on dev_get_flags()

        Returns:
            List[str]: A list of flag names
        """
        choices = self._get_flag_choices()
        clear_flags = choices.get("IFF_PROMISC", 0)
        clear_flags |= choices.get("IFF_ALLMULTI", 0)
        clear_flags |= choices.get("IFF_RUNNING", 0)
        clear_flags |= choices.get("IFF_LOWER_UP", 0)
        clear_flags |= choices.get("IFF_DORMANT", 0)

        clear_gflags = choices.get("IFF_PROMISC", 0)
        clear_gflags |= choices.get("IFF_ALLMULTI)", 0)

        flags = (self.flags & ~clear_flags) | (self.gflags & ~clear_gflags)

        if self.is_running():
            if self.is_operational():
                flags |= choices.get("IFF_RUNNING", 0)
            if self.is_carrier_ok():
                flags |= choices.get("IFF_LOWER_UP", 0)
            if self.is_dormant():
                flags |= choices.get("IFF_DORMANT", 0)

        net_device_flags_enum_flags = wrappers.Flags(choices)
        net_device_flags = net_device_flags_enum_flags(flags)

        # It's preferable to provide a deterministic list of items. i.e. for testing
        return sorted(net_device_flags)

    @property
    def promisc(self) -> bool:
        """Return if this network interface is in promiscuous mode.

        Returns:
            bool: True if this network interface is in promiscuous mode. Otherwise, False
        """
        return self.flags & self._get_net_device_flag_value("IFF_PROMISC") != 0

    def _do_get_net_namespace_id(self) -> int:
        """Return the network namespace id for this network interface.

        Returns:
            int: the network namespace id for this network interface
        """
        nd_net = self.nd_net
        if nd_net.has_member("net"):
            # In kernel 4.1.52 the 'nd_net' member type was changed from
            # 'struct net*' to 'possible_net_t' which has a 'struct net *net' member.
            net_ns_id = nd_net.net.get_inode()
        else:
            # In kernels < 4.1.52 the 'nd_net'member type was  'struct net*'
            net_ns_id = nd_net.get_inode()

        return net_ns_id

    def get_net_namespace_id(self) -> Optional[int]:
        """Return the network namespace id for this network interface.

        Returns:
            int: the network namespace id for this network interface
        """
        try:
            return self._do_get_net_namespace_id()
        except exceptions.InvalidAddressException:
            vollog.debug(
                f"Encountered an invalid address exception when getting the namespace for {self.vol.offset:#x}"
            )
            return None

    def get_operational_state(self) -> Union[str, interfaces.renderers.BaseAbsentValue]:
        """Return the netwok device oprational state (RFC 2863) string

        Returns:
            str: A string with the operational state
        """
        try:
            return linux_constants.IF_OPER_STATES(self.operstate).name
        except ValueError:
            vollog.warning(f"Invalid net_device operational state '{self.operstate}'")
            return renderers.UnparsableValue()

    def get_qdisc_name(self) -> Optional[str]:
        """Return the network device queuing discipline (qdisc) name

        Returns:
            str: A string with the queuing discipline (qdisc) name
        """
        try:
            return utility.array_to_string(self.qdisc.ops.id)
        except exceptions.InvalidAddressException:
            vollog.debug(f"Unable to get qdisc name for {self.vol.offset:#x}")
            return None

    def get_queue_length(self) -> int:
        """Return the network device transmission queue length (qlen)

        Returns:
            int: the network device transmission queue length (qlen)
        """
        return self.tx_queue_len


class in_device(objects.StructType):
    def get_addresses(
        self, max_devices=128
    ) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        """Yield the IPv4 ifaddr addresses

        Yields:
            in_ifaddr: An IPv4 ifaddr address
        """
        seen = set()

        try:
            cur = self.ifa_list
        except exceptions.InvalidAddressException:
            return

        while cur and cur.vol.offset:
            if len(seen) > max_devices:
                break

            if cur.vol.offset in seen:
                break
            seen.add(cur.vol.offset)

            yield cur

            try:
                cur = cur.ifa_next
            except exceptions.InvalidAddressException:
                break


class inet6_dev(objects.StructType):
    def get_addresses(
        self,
    ) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        """Yield the IPv6 ifaddr addresses

        Yields:
            inet6_ifaddr: An IPv6 ifaddr address
        """
        if not self.has_member(
            "addr_list"
        ) or not self.addr_list.vol.type_name.endswith(constants.BANG + "list_head"):
            # kernels < 3.0
            # FIXME: struct inet6_ifaddr	*addr_list;
            vollog.warning(
                "IPv6 is unsupported for this kernel. Check if the ISF contains the appropriate 'inet6_dev' type"
            )
            return

        symbol_space = self._context.symbol_space
        table_name = self.get_symbol_table_name()
        inet6_ifaddr_symname = table_name + constants.BANG + "inet6_ifaddr"
        if not symbol_space.has_type(inet6_ifaddr_symname) or not symbol_space.get_type(
            inet6_ifaddr_symname
        ).has_member("if_list"):
            vollog.warning(
                "IPv6 is unsupported for this kernel. Check if the ISF contains the appropriate 'inet6_ifaddr' type"
            )
            return

        # 'if_list' member was added to 'inet6_ifaddr' type in kernels 3.0
        yield from self.addr_list.to_list(inet6_ifaddr_symname, "if_list")


class in_ifaddr(objects.StructType):
    # Translation to text based on iproute2 package. See 'rtnl_rtscope_tab' in lib/rt_names.c
    _rtnl_rtscope_tab = {
        "RT_SCOPE_UNIVERSE": "global",
        "RT_SCOPE_NOWHERE": "nowhere",
        "RT_SCOPE_HOST": "host",
        "RT_SCOPE_LINK": "link",
        "RT_SCOPE_SITE": "site",
    }

    def get_scope_type(self) -> str:
        """Get the scope type for this IPv4 address

        Returns:
            str: the IPv4 scope type.
        """
        table_name = self.get_symbol_table_name()
        rt_scope_enum = self._context.symbol_space.get_enumeration(
            table_name + constants.BANG + "rt_scope_t"
        )
        try:
            rt_scope = rt_scope_enum.lookup(self.ifa_scope)
        except ValueError:
            return "unknown"

        return self._rtnl_rtscope_tab.get(rt_scope, "unknown")

    def get_address(self) -> str:
        """Get an string with the IPv4 address

        Returns:
            str: the IPv4 address
        """
        return conversion.convert_ipv4(self.ifa_address)

    def get_prefix_len(self) -> int:
        """Get the IPv4 address prefix len

        Returns:
            int: the IPv4 address prefix len
        """
        return self.ifa_prefixlen


class inet6_ifaddr(objects.StructType):
    def get_scope_type(self) -> str:
        """Get the scope type for this IPv6 address

        Returns:
            str: the IPv6 scope type.
        """
        if (self.scope & linux_constants.IFA_HOST) != 0:
            return "host"
        elif (self.scope & linux_constants.IFA_LINK) != 0:
            return "link"
        elif (self.scope & linux_constants.IFA_SITE) != 0:
            return "site"
        else:
            return "global"

    def get_address(self) -> str:
        """Get an string with the IPv6 address

        Returns:
            str: the IPv6 address
        """
        return conversion.convert_ipv6(self.addr.in6_u.u6_addr32)

    def get_prefix_len(self) -> int:
        """Get the IPv6 address prefix len

        Returns:
            int: the IPv6 address prefix len
        """
        return self.prefix_len


class socket(objects.StructType):
    def _get_vol_kernel(self) -> interfaces.context.ModuleInterface:
        symbol_table_arr = self.vol.type_name.split("!", 1)
        symbol_table = symbol_table_arr[0] if len(symbol_table_arr) == 2 else None

        if symbol_table is None:
            raise ValueError(f"No module using the symbol table {symbol_table}")

        module_names = list(
            self._context.modules.get_modules_by_symbol_tables(symbol_table)
        )
        if not module_names:
            raise ValueError(f"No module using the symbol table {symbol_table}")
        kernel_module_name = module_names[0]
        kernel = self._context.modules[kernel_module_name]
        return kernel

    def get_inode(self) -> int:
        try:
            kernel = self._get_vol_kernel()
        except ValueError:
            return 0
        socket_alloc = linux.LinuxUtilities.container_of(
            self.vol.offset, "socket_alloc", "socket", kernel
        )
        if socket_alloc is None:
            return 0
        vfs_inode = socket_alloc.vfs_inode

        return vfs_inode.i_ino

    def get_state(self) -> str:
        socket_state_idx = self.state
        if 0 <= socket_state_idx < len(linux_constants.SOCKET_STATES):
            return linux_constants.SOCKET_STATES[socket_state_idx]
        return "Unknown socket state"


class sock(objects.StructType):
    def get_family(self) -> str:
        family_idx = self.__sk_common.skc_family
        if 0 <= family_idx < len(linux_constants.SOCK_FAMILY):
            return linux_constants.SOCK_FAMILY[family_idx]
        return "Unknown socket family"

    def get_type(self) -> str:
        return linux_constants.SOCK_TYPES.get(self.sk_type, "")

    def get_inode(self) -> int:
        if not self.sk_socket:
            return 0
        return self.sk_socket.get_inode()

    def get_protocol(self) -> Optional[str]:
        return None

    def get_state(self) -> str:
        # Return the generic socket state
        if self.has_member("sk"):
            return self.sk.sk_socket.get_state()
        return self.sk_socket.get_state()


class unix_sock(objects.StructType):
    def get_name(self) -> Optional[str]:
        if not self.addr:
            return None
        sockaddr_un = self.addr.name.cast("sockaddr_un")
        saddr = str(utility.array_to_string(sockaddr_un.sun_path))
        return saddr

    def get_protocol(self) -> Optional[str]:
        return None

    def get_state(self) -> str:
        """Return a string representing the sock state."""

        # Unix socket states reuse (a subset) of the inet_sock states contants
        if self.sk.get_type() == "STREAM":
            state_idx = self.sk.__sk_common.skc_state
            if 0 <= state_idx < len(linux_constants.TCP_STATES):
                return linux_constants.TCP_STATES[state_idx]
            else:
                return "Unknown unix_sock stream state"
        # Return the generic socket state
        return self.sk.sk_socket.get_state()

    def get_inode(self) -> int:
        return self.sk.get_inode()


class inet_sock(objects.StructType):
    def get_family(self) -> str:
        family_idx = self.sk.__sk_common.skc_family
        if 0 <= family_idx < len(linux_constants.SOCK_FAMILY):
            return linux_constants.SOCK_FAMILY[family_idx]
        return "Unknown inet_sock family"

    def get_protocol(self) -> Optional[str]:
        # If INET6 family and a proto is defined, we use that specific IPv6 protocol.
        # Otherwise, we use the standard IP protocol.
        protocol = linux_constants.IP_PROTOCOLS.get(self.sk.sk_protocol)
        if self.get_family() == "AF_INET6":
            protocol = linux_constants.IPV6_PROTOCOLS.get(self.sk.sk_protocol, protocol)
        return protocol

    def get_state(self) -> str:
        """Return a string representing the sock state."""

        if self.sk.get_type() == "STREAM":
            state_idx = self.sk.__sk_common.skc_state
            if 0 <= state_idx < len(linux_constants.TCP_STATES):
                return linux_constants.TCP_STATES[state_idx]
            else:
                return "Unknown inet_sock stream state"
        # Return the generic socket state
        return self.sk.sk_socket.get_state()

    def get_src_port(self) -> Optional[int]:
        sport_le = getattr(self, "sport", getattr(self, "inet_sport", None))
        if sport_le is not None:
            return socket_module.htons(sport_le)
        return None

    def get_dst_port(self) -> Optional[int]:
        sk_common = self.sk.__sk_common
        if hasattr(sk_common, "skc_portpair"):
            dport_le = sk_common.skc_portpair & 0xFFFF
        elif hasattr(self, "dport"):
            dport_le = self.dport
        elif hasattr(self, "inet_dport"):
            dport_le = self.inet_dport
        elif hasattr(sk_common, "skc_dport"):
            dport_le = sk_common.skc_dport
        else:
            return None
        return socket_module.htons(dport_le)

    def get_src_addr(self) -> Optional[str]:
        sk_common = self.sk.__sk_common
        family = sk_common.skc_family
        if family == socket_module.AF_INET:
            addr_size = 4
            if hasattr(self, "rcv_saddr"):
                saddr = self.rcv_saddr
            elif hasattr(self, "inet_rcv_saddr"):
                saddr = self.inet_rcv_saddr
            else:
                saddr = sk_common.skc_rcv_saddr
        elif family == socket_module.AF_INET6:
            addr_size = 16
            saddr = self.pinet6.saddr
        else:
            return None
        parent_layer = self._context.layers[self.vol.layer_name]
        try:
            addr_bytes = parent_layer.read(saddr.vol.offset, addr_size)
        except exceptions.InvalidAddressException:
            vollog.debug(
                f"Unable to read socket src address from {saddr.vol.offset:#x}"
            )
            return None
        return socket_module.inet_ntop(family, addr_bytes)

    def get_dst_addr(self) -> Optional[str]:
        sk_common = self.sk.__sk_common
        family = sk_common.skc_family
        if family == socket_module.AF_INET:
            if hasattr(self, "daddr") and self.daddr:
                daddr = self.daddr
            elif hasattr(self, "inet_daddr") and self.inet_daddr:
                daddr = self.inet_daddr
            else:
                daddr = sk_common.skc_daddr
            addr_size = 4
        elif family == socket_module.AF_INET6:
            if hasattr(self.pinet6, "daddr"):
                daddr = self.pinet6.daddr
            else:
                daddr = sk_common.skc_v6_daddr
            addr_size = 16
        else:
            return None
        parent_layer = self._context.layers[self.vol.layer_name]
        try:
            addr_bytes = parent_layer.read(daddr.vol.offset, addr_size)
        except exceptions.InvalidAddressException:
            vollog.debug(
                f"Unable to read socket dst address from {daddr.vol.offset:#x}"
            )
            return None
        return socket_module.inet_ntop(family, addr_bytes)


class netlink_sock(objects.StructType):
    def get_protocol(self) -> str:
        protocol_idx = self.sk.sk_protocol
        if 0 <= protocol_idx < len(linux_constants.NETLINK_PROTOCOLS):
            return linux_constants.NETLINK_PROTOCOLS[protocol_idx]
        return "Unknown netlink_sock protocol"

    def get_state(self):
        # Return the generic socket state
        return self.sk.sk_socket.get_state()

    def get_portid(self) -> int:
        if self.has_member("pid"):
            # kernel < 3.7.10
            return self.pid
        if self.has_member("portid"):
            # kernel >= 3.7.10
            return self.portid
        else:
            raise AttributeError("Unable to find a source port id")

    def get_dst_portid(self) -> int:
        if self.has_member("dst_pid"):
            # kernel < 3.7.10
            return self.dst_pid
        if self.has_member("dst_portid"):
            # kernel >= 3.7.10
            return self.dst_portid
        else:
            raise AttributeError("Unable to find a destination port id")


class vsock_sock(objects.StructType):
    def get_protocol(self):
        # The protocol should always be 0 for vsocks
        return None

    def get_state(self):
        # Return the generic socket state
        return self.sk.sk_socket.get_state()


class packet_sock(objects.StructType):
    def get_protocol(self) -> Optional[str]:
        eth_proto = socket_module.htons(self.num)
        if eth_proto == 0:
            return None
        elif eth_proto in linux_constants.ETH_PROTOCOLS:
            return linux_constants.ETH_PROTOCOLS[eth_proto]
        else:
            return f"0x{eth_proto:x}"

    def get_state(self):
        # Return the generic socket state
        return self.sk.sk_socket.get_state()


class bt_sock(objects.StructType):
    def get_protocol(self) -> Optional[str]:
        type_idx = self.sk.sk_protocol
        if 0 <= type_idx < len(linux_constants.BLUETOOTH_PROTOCOLS):
            return linux_constants.BLUETOOTH_PROTOCOLS[type_idx]
        return None

    def get_state(self) -> Optional[str]:
        state_idx = self.sk.__sk_common.skc_state
        if 0 <= state_idx < len(linux_constants.BLUETOOTH_STATES):
            return linux_constants.BLUETOOTH_STATES[state_idx]
        return None


class xdp_sock(objects.StructType):
    def get_protocol(self):
        # The protocol should always be 0 for xdp_sock
        return None

    def get_state(self):
        # xdp_sock.state is an enum
        return self.state.lookup()



# This file is Copyright 2020 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Dict, Generator, List

import volatility3.framework.symbols.linux.utilities.modules as linux_utilities_modules
from volatility3.framework import constants, deprecation, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.symbols.linux import extensions

vollog = logging.getLogger(__name__)


class Check_modules(interfaces.plugins.PluginInterface):
    """Compares module list to sysfs info, if available"""

    _version = (3, 0, 1)
    _required_framework_version = (2, 0, 0)

    @classmethod
    def compare_kset_and_lsmod(
        cls, context: interfaces.context.ContextInterface, vmlinux_name: str
    ) -> Generator[extensions.module, None, None]:
        kset_modules = linux_utilities_modules.Modules.get_kset_modules(
            context=context, vmlinux_name=vmlinux_name
        )

        lsmod_modules = set(
            str(utility.array_to_string(modules.name))
            for modules in linux_utilities_modules.Modules.list_modules(
                context=context, vmlinux_module_name=vmlinux_name
            )
        )

        for mod_name in set(kset_modules.keys()).difference(lsmod_modules):
            yield kset_modules[mod_name]

    implementation = compare_kset_and_lsmod

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=constants.architectures.LINUX_ARCHS,
            ),
            requirements.VersionRequirement(
                name="modules",
                component=linux_utilities_modules.Modules,
                version=(3, 0, 1),
            ),
            requirements.VersionRequirement(
                name="linux_utilities_modules_module_display_plugin",
                component=linux_utilities_modules.ModuleDisplayPlugin,
                version=(2, 0, 0),
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract listed modules",
                default=False,
                optional=True,
            ),
        ]

    @classmethod
    @deprecation.deprecated_method(
        replacement=linux_utilities_modules.Modules.get_kset_modules,
        removal_date="2026-03-25",
        replacement_version=(3, 0, 0),
    )
    def get_kset_modules(
        cls, context: interfaces.context.ContextInterface, vmlinux_name: str
    ) -> Dict[str, extensions.module]:
        return linux_utilities_modules.Modules.get_kset_modules(context, vmlinux_name)

    def run(self):
        return renderers.TreeGrid(
            linux_utilities_modules.ModuleDisplayPlugin.columns_results,
            self._generator(),
        )

    def _generator(self):
        yield from linux_utilities_modules.ModuleDisplayPlugin.generate_results(
            self.context,
            self.implementation,
            self.config["kernel"],
            self.config["dump"],
            self.open,
        )
